import abc
from fireworks.core import Message, Junction
from fireworks.utils.exceptions import EndHyperparameterOptimization
from .database import create_table, TablePipe
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column
from sqlalchemy_utils import JSONType as JSON
from collections import defaultdict
import types

def update(bundle: dict, parameters: dict):
    """
    Args:
        bundle - A dictionary of key: (obj, atr). Obj is the object referred to, and attr is a string with the name of the attribute to be assigned.
        parameters - A dictionary of key: value. Wherever keys match, obj.attr will be set to value.
    """
    for key, param in parameters.items():
        if key in bundle:
            obj, atr = bundle[key]
            setattr(obj, attr, param)

class Factory(Junction):
    """
    Base class for hyperparameter optimization in pytorch using queues.
    """
    # NOTE: This is currently not parallelized. It would be nice if it was.

    required_components = {'trainer': types.FunctionType, 'eval_set': object, 'parameterizer': types.FunctionType, 'metrics': dict}

    def __init__(self, *args, components=None, **kwargs):

        Junction.__init__(self, *args, components=components, **kwargs)
        self.get_connection()

    @abc.abstractmethod
    def get_connection(self): pass

    def run(self):
        while True:
            past_params, past_metrics = self.read()
            try:
                # Generate new set of parameters
                params = self.parameterizer(past_params, past_metrics)
                # Generate an evaluator
                evaluator = self.trainer(params)
                # NOTE: This part is pytorch ignite syntax
                for name, metric in self.metrics.items():
                    metric.attach(evaluator, name) # TODO: Make sure this resets the metric
                # Running the evaluator should perform training on the dataset followed by evlaution and return evaluation metrics
                evaluator.run(self.eval_set, max_epochs=1)
                # Evaluate the metrics that were attached to the evaluator
                computed_metrics = {name: metric.compute() for name, metric in self.metrics.items()}
                self.write(params, computed_metrics)
                evaluator = None
            except EndHyperparameterOptimization:
                self.after()
                break

    @abc.abstractmethod
    def read(self): pass

    @abc.abstractmethod
    def write(self, params, metrics_dict): pass

    def train(self, params):

        self.trainer.model.update_components(params)

    def after(self, *args, **kwargs): pass

class LocalMemoryFactory(Factory):
    """
    Factory that stores parameters in memory.
    """

    def get_connection(self):
        self.params = Message()
        self.computed_metrics = defaultdict(Message)

    def read(self):
        return self.params, self.computed_metrics

    def write(self, params, metrics_dict):
        self.params = self.params.append(params)
        for key in metrics_dict:
            self.computed_metrics[key] = self.computed_metrics[key].append(metrics_dict[key])

class SQLFactory(Factory):
    """
    Factory that stores parameters in SQLalchemy database while caching them locally.
    """
    required_components = {
        'trainer': types.FunctionType,
        'eval_set': object,
        'parameterizer': types.FunctionType,
        'metrics': dict,
        'engine': object,
        'params_table': object,
        'metrics_tables': object,
        }

    def __init__(self,*args, components=None, **kwargs):

        Junction.__init__(self, *args, components=components, **kwargs)
        self.params_pipe = TablePipe(self.params_table, self.engine)
        self.metrics_pipes = {key: TablePipe(value, self.engine) for key, value in self.metrics_tables.items()}
        self.computed_metrics = defaultdict(Message)
        self.get_connection()

    def get_connection(self):

        # TODO: Ensure id consistency accross these tables using foreign key constraints. This should implicitly
        # hold true without such constraints however, because these tables are updated in sync.
        for table in self.metrics_tables.values():
            table.metadata.create_all(self.engine)
        self.params_table.metadata.create_all(self.engine)
        self.id = 0
        self.sync()

    def write(self, params, metrics):

        # if len(params) != len(metrics):
        #     raise ValueError("Parameters and Metrics messages must be equal length.")
        params = Message(params)
        for key, metric in metrics.items():
            self.computed_metrics[key] = self.computed_metrics[key].append(metric)
            self.metrics_pipes[key].insert(metric)
            self.metrics_pipes[key].commit()
        self.params = self.params.append(params)
        self.params_pipe.insert(params)
        self.params_pipe.commit()

    def read(self):

        return self.params, self.computed_metrics

    def read_db(self):

        return self.params_pipe.query().all(), {key: pipe.query().all() for key, pipe in self.metrics_pipes.items()}

    def sync(self):
        """ Syncs local copy of metrics and params with db. """
        self.params, self.computed_metrics = self.read_db()

    def after(self):
        self.sync()



import fireworks
from fireworks import Message
from fireworks.extensions import IgniteJunction, Experiment
from ignite.metrics import Metric

import torch
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from nonlinear_regression_utils import NonlinearModel, get_data
# from examples.database_example import get_data

description = "In this experiment, we are training a polynomial model using least squares regression to fit data generated by a random polynomial."
experiment = Experiment("nonlinear_regression", description=description)

train_set, test_set, params = get_data(n=1000)

model = NonlinearModel()
# model = NonlinearModel(components={'d': [0], 'e':[0]})
# model.freeze(['d','e'])

# Construct training closure and train using ignite
base_loss = torch.nn.MSELoss()
loss = lambda batch: base_loss(batch['y_pred'], batch['y'])
trainer = IgniteJunction(components={'model': model, 'dataset': train_set}, loss=loss, optimizer='Adam', lr=.1, visdom=False)

class ModelSaverMetric(Metric):

    def __init__(self, output_transform=lambda x:x, log_interval=100):
        self.model_state = Message()
        Metric.__init__(self, output_transform=output_transform)
        self.log_interval = log_interval

    def iteration_completed(self, engine):
            iter = (engine.state.iteration-1)
            if iter % self.log_interval == 0:
                current_state = Message.from_objects(deepcopy(engine.state.output['state']))
                current_state['iteration'] = [iter]
                self.model_state = self.model_state.append(current_state)

    def compute(self):
        # Return most recent model state
        l = len(self.model_state)
        return self.model_state[l-1]

    def reset(self):
        pass
    
    def update(self, output):
        pass

if __name__== "__main__":

    model_state_metric = ModelSaverMetric()
    model_state_metric.attach(trainer, 'state')

    x = Message({'x':np.arange(-10,10,.2)}).to_tensors()

    # Run initial evaluation
    y_initial = model(x)['y_pred'].detach().numpy()
    initial_loss = loss(model(test_set[0:250]))
    print("Initial loss on test set: {0}".format(initial_loss))

    # Save initial state of model
    file_path = experiment.open('initial_model', string_only=True)
    initial_state = model.get_state()
    Message.from_objects(initial_state).to('json', path=file_path)

    trainer.train(max_epochs=200)

    final_loss = loss(model(test_set[0:250]))
    print("Final loss on test set:: {0}".format(final_loss))

    # Visualize functions
    true_model = NonlinearModel(components={'a':[params['a']], 'b': [params['b']], 'c': [params['c']], 'd': [0], 'e': [0]})

    y_true = true_model(x)['y_pred'].detach().numpy()
    y_final = model(x)['y_pred'].detach().numpy()

    # Save model states during training
    file_path = experiment.open('model_states', string_only=True)
    model_states = model_state_metric.model_state
    Message.from_objects(initial_state).to('json', path=file_path)

    fig, ax = plt.subplots()

    def animate(frame):

        current_state = {'internal': frame['internal'][0], 'external': {}}
        model.set_state(current_state)

        y_predicted = model(x)['y_pred'].detach().numpy()
        xdata = list(x['x'].detach().numpy())
        ydata = list(y_predicted)
        ax.clear()
        ax.plot(xdata, list(y_true), 'r')
        ax.plot(xdata, ydata, 'g')
        title = "Iteration: {0}".format(frame['iteration'][0])

        ax.set_title(title)

    ani = FuncAnimation(fig, animate, model_state_metric.model_state, interval=1000)
    ani.save(experiment.open("models.mp4", string_only=True)) # This will only work if you have ffmpeg installed.
    plt.show()



import fireworks
import os
import pandas as pd
from fireworks.toolbox import pipes as pl
from fireworks import Message
from fireworks.utils import index_to_list
from fireworks.utils.test_helpers import *
import torch
import numpy as np
import itertools

test_dir = fireworks.test_dir

def conforms_to_spec(pipe):

    assert hasattr(pipe, '__iter__')
    assert hasattr(pipe, '__next__')

    return True

def test_pipe(): pass

def test_Title2LabelPipe():

    dumbo = one_way_dummy()
    rumbo = one_way_dummy()
    bumbo = reset_dummy()
    gumbo = next_dummy()
    jumbo = next_dummy()

    labeler = pl.Title2LabelPipe( 'yes', dumbo)
    mislabeler = pl.Title2LabelPipe( 'yes', rumbo, labels_column='barrels')
    donkeykong = pl.Title2LabelPipe('yes', gumbo)
    dixiekong = pl.Title2LabelPipe( 'yes', jumbo, labels_column='bananas')
    labeler.reset()
    assert labeler.count == 0
    mislabeler.reset()
    assert mislabeler.count == 0
    def test_iteration(pipe, n, label):
        for i in range(n):
            x = pipe.__next__()
            assert len(x) == 1
            assert (x[label] == ['yes']).all()

    test_iteration(labeler, 10, 'labels')
    test_iteration(mislabeler, 10, 'barrels')
    test_iteration(donkeykong, 12, 'labels')
    test_iteration(dixiekong, 14, 'bananas')
    assert labeler.count == 10
    assert mislabeler.count == 10
    assert donkeykong.count == 12
    assert dixiekong.count == 14

    labeler = pl.Title2LabelPipe('yes', bumbo)
    labeler.reset()
    assert labeler.count == 0



def test_BioSeqPipe():

    test_file = os.path.join(test_dir, 'sample_genes.fa')
    genes = pl.BioSeqPipe(test_file)
    assert conforms_to_spec(genes)
    f = lambda batch: [1 for _ in batch]
    embedding_function = {'sequences': f}

    for gene in genes:
        assert set(['sequences', 'ids', 'names', 'descriptions', 'dbxrefs']) == set(gene.columns)
        assert type(gene) is Message
        assert set(gene.tensor_message.keys()) == set()
        assert set(gene.df.keys()) == set(['ids', 'names', 'descriptions', 'dbxrefs', 'sequences'])
        assert len(gene) == 1
    # Reset and do it again to confirm we can repeat the Pipe
    genes.reset()
    for gene in genes:
        assert set(['sequences', 'ids', 'names', 'descriptions', 'dbxrefs']) == set(gene.columns)
        assert type(gene) is Message
        assert set(gene.tensor_message.keys()) == set()
        assert set(gene.df.keys()) == set(['ids', 'names', 'descriptions', 'dbxrefs', 'sequences'])
        assert len(gene) == 1

def test_LoopingPipe():

    dumbo = one_way_dummy()
    loopy = pl.LoopingPipe(dumbo)
    loopy[10]
    loopy[5]
    numba1 = len(loopy)
    numba2 = len(loopy) # Call __len__ twice to ensure both compute_length and retrieval of length works
    assert numba1 == 20
    assert numba2 == 20
    x = loopy[0]
    assert x == Message({'count': [0]})
    loopy[10]
    loopy = pl.LoopingPipe(dumbo)
    x = loopy[0]
    assert x == Message({'count': [0]}) # Check that the input Pipes were reset.
    assert loopy.length is None
    try: # Test if length is implicitly calculated whenever input Pipes run out.
        loopy[21]
    except IndexError:
        assert True
    else:
        assert False
    assert loopy.length == 20

def test_RepeaterPipe():

    dumbo = one_way_iter_dummy()
    robert = pl.RepeaterPipe(dumbo)
    numbaz = Message()
    assert len(numbaz) == 0
    for numba in robert:
        numbaz = numbaz.append(numba)
    assert len(numbaz) == robert.repetitions*20

    dumbo = one_way_dummy()
    robert = pl.RepeaterPipe(dumbo)
    numbaz = Message()
    robert.reset()
    i = 0
    assert len(numbaz) == 0
    while True:
        try:
            numbaz = numbaz.append(robert.__next__())
            i+=1
            if i > 1000: # If something goes horribly wrong, cancel test
                assert False
        except StopIteration:
            break
    assert len(numbaz) == robert.repetitions*20

def test_CachingPipe():

    # Test type checking
    dumbo = getitem_dummy()
    cashmoney = pl.CachingPipe(dumbo)

    # Test __getitem__
    assert cashmoney.cache.cache == Message()
    x = cashmoney[2]
    assert x == Message({'values':[2]})
    assert cashmoney.cache.cache == Message({'values':[2]})
    x = cashmoney[2,3,4]
    x = cashmoney[2,3,4]
    assert x == Message({'values':[2,3,4]})
    assert cashmoney.cache.cache == Message({'values':[2,3,4]})
    x = cashmoney[5,2,3]
    assert x == Message({'values':[5,2,3]})
    assert set(cashmoney.cache.pointers.keys()) == set([2,3,4,5])
    x = cashmoney[8,5,1,4,0,9,2]
    assert x == Message({'values':[8,5,1,4,0,9,2]})
    assert set(cashmoney.cache.pointers.keys()) == set([0,1,2,3,4,5,8,9])
    # Test explicit length calculation
    assert cashmoney.length is None
    length = len(cashmoney)
    assert length == 20
    # Test implicit length calculation
    cashmoney = pl.CachingPipe(dumbo)
    assert cashmoney.length is None
    assert cashmoney.lower_bound == 0
    cashmoney[10]
    assert cashmoney.lower_bound == 10
    cashmoney[8]
    assert cashmoney.lower_bound == 10
    cashmoney[12,15]
    assert cashmoney.lower_bound == 15
    length = len(cashmoney)
    assert length == 20

def test_IndexMapperPipe():

    getty = getitem_dummy()
    input_indices = [i for i in range(12)]
    output_indices = [i for i in reversed(range(12))]

    flipped = pl.IndexMapperPipe(input_indices, output_indices,getty)
    assert len(flipped) == 12
    for i in range(12):
        assert (flipped[i]['values'] == 11-i).all()

def test_ShufflingPipe():

    bobby = getitem_dummy()
    shuffla = pl.ShufflerPipe(input=bobby)
    shuffla[2]
    shuffled = False
    for shu, i in zip(shuffla, itertools.count()):
        if shu['values'][0] != i:
            shuffled = True
    assert shuffled

def test_FunctionPipe():

    bobby = smart_dummy()
    f = lambda m: m.append(m)
    functron = pl.FunctionPipe(input=bobby, function=f)
    # Test call
    single = bobby()
    double = functron()

    # Test getitem
    single = bobby[3:5]
    double = functron[3:5]
    assert double == single.append(single)
    # Test iteration
    sobby = smart_dummy()
    for single, double in zip(sobby, functron): # We can't iterate through bobby, since that would increment functron as well
        assert double == single.append(single)

def test_TensorPipe():

    bobby = smart_dummy()
    tensors = pl.TensorPipe(bobby)
    assert type(tensors[2]['values']) is torch.Tensor
    assert type(bobby[2]['values']) is pd.Series
    for x in tensors:
        assert type(x['values']) is torch.Tensor



from .message import Message
from abc import ABC, abstractmethod
import pandas as pd
from itertools import count
from bidict import bidict
import numpy as np

class MessageCache(ABC):
    """
    A message cache stores parts of a larger method and supports retrievals and
    insertions based on index.
    The use case for a MessageCache is for storing parts of a large dataset in memory.
    The MessageCache can keep track of which elements and which indices are present
    in memory at a given time and allow for updates and retrievals.
    """

    def __init__(self, max_size):
        """
        Args:
            max_size (int): The maximum size of this cache.
        """
        self.max_size = max_size
        self.cache = Message()
        self.pointers = bidict()

    def __getitem__(self, index): # TODO: Optimize for range queries.
        """
        Translates the given index into the appropriate index for the internal message and returns those values.
        """
        if type(index) is slice:
            index = slice_to_list(index)

        if type(index) is int:
            pointers = self.pointers[index]
        else:
            pointers = [self.pointers[i] for i in index] # This will raise an error if index not present

        return self.cache[pointers] #list(pointers.keys())]#.tolist()]

    @abstractmethod
    def __setitem__(self, index, message): pass

    def __getattr__(self, *args, **kwargs):
        return self.cache.__getattr__(args, kwargs)

    def __contains__(self, item):

        return item in self.pointers.keys()

    def __len__(self):
        return len(self.cache)

    def __repr__(self):
        return "MessageCache with indices {0}.".format(list(self.pointers.keys()))#.tolist())

    def insert(self, index, message):
        """
        Inserts message into cache along with the desired indices.
        This method should be called by __setitem__ as needed to perform the insertion.

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.
        """
        message = Message(message)

        if type(index) is int:
            index = [index]
        if type(index) is slice:
            index = slice_to_list(index)
        present = [i for i in index if i in self.pointers.keys()] # Intersection
        not_present = [i for i in index if i not in self.pointers.keys()] # Difference
        # Get indices in message
        present_indices = get_indices(index, present)
        not_present_indices = get_indices(index, not_present)

        self._add_new(not_present, message[not_present_indices])
        self._update_existing(present, message[present_indices])

    def _add_new(self, index, message):
        """
        Adds new elements to cache and updates pointers.
        This method should be called by __setitem__ or insert as needed to perform insertions.

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.

        """
        start = len(self.cache)
        self.cache = self.cache.append(message)
        stop = len(self.cache)
        for i,j in zip(range(start, stop), index):
            self.pointers[j] = i

    def _update_existing(self, index, message):
        """
        Updates elements already in the message.
        This method should be called by __setitem__ or insert as needed to perform updates.

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.

        """
        if index:
            if type(index) is int:
                indices = self.pointers[index]
            else:
                indices = [self.pointers[i] for i in index]
            self.cache[indices] = message #indices.tolist()

    def delete(self, index):
        """
        Deletes elements in the message corresponding to index.
        This method should be called by __setitem__ or __delitem__ as needed.

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.

        """
        # Get the index in the internal cache corresponding to the virtual (argument) index
        if type(index) is slice:
            index = slice_to_list(index)
        if type(index) is int:
            index = [index]

        index = [i for i in index if i in self.pointers]
        internal_indices = [self.pointers[i] for i in index]
        # Delete those elements
        del self.cache[internal_indices]
        # Delete the pointers for those elements
        for i in index:
            del self.pointers[i]
        f = pointer_adjustment_function(internal_indices)
        self.pointers = dict(self.pointers) # Convert to normal dict to prevent duplication errors
        for key, value in self.pointers.items():
            self.pointers[key] -= f(value)
        self.pointers = bidict(self.pointers)

    def _permute(self, permutation, target = None):
        """
        Rearranges elements of cache based on permutation.
        Permutation should be a list of indices. Can optionally specify target, which is the indices to apply the permutation to.
        By default, the entire index will be treated as the target.

        Args:
            permutation: A list of indices
        """
        if target is None:
            target = [self.pointers.inv[i] for i in range(len(self.cache))]

        pointers = bidict({t:self.pointers[self.pointers.inv[p]] for t,p in zip(target, permutation)}) # Permute which internal index a pointer points to
        index = [self.pointers[t] for t in target]
        self.cache[index] = self.cache[permutation]  # Permute internal indices
        self.pointers = pointers

    def sort(self):
        """
        Rearranges internal cache indices to be in sorted order.
        """
        # Identify sort permutation
        # Apply sort permutation
        pass

    def search(self, **kwargs):
        pass

    @property
    def size(self):
        return len(self.cache)

class UnlimitedCache(MessageCache):
    """
    This is a basic implementation of a MessageCache that simply appends new
    elements and never clears memory internally
    """

    def __init__(self):
        super().__init__(None) # No max size

    def __setitem__(self, index, message):
        """
        Simply inserts the message with the corresponding index without ever freeing memory.

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.

        """
        self.insert(index, message)

    def __delitem__(self, index):
        """
        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.

        """
        self.delete(index)

class BufferedCache(MessageCache):
    """
    This implements a setitem method that assumes that when the cache is full, elements must be deleted until it is max_size - buffer_size
    in length. The deletion method, _free, must be implemented by a subclass.
    """


    def init_buffer(self, buffer_size = 0):

        self.buffer_size = buffer_size # Specifies how far above max_size the cache can go before having to delete elements

    def __setitem__(self, index, message):
        """

        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.

        """
        index = index_to_list(index)
        message = Message(message)
        if len(index) != len(message):
            raise ValueError("Message length does not match length of index for insertion.")
        # Determine how much space to free in order to insert message
        free_space = self.max_size - len(self)

        if len(message) > free_space:
            how_much = max(self.buffer_size - free_space, len(message) - free_space)

            self.free(how_much)

        self.insert(index, message)

    def __delitem__(self, index):
        """
        Args:
            index: The index to insert into. Can be an int, slice, or list of integer indices.
            message: The Message to insert. Should have the same length as the provided idnex.

        """
        self.delete(index)

    @abstractmethod
    def free(self, n): pass

class RRCache(BufferedCache):

    def free(self, n):
        indices_present = self.cache.keys()
        delete_indices = list(np.random.choice(indices_present, n, replace=False))
        self.__delete__(delete_indices)

class RankingCache(MessageCache):
    """
    Implements a free method that deletes elements based on a ranking function.
    """

    def init_rank_dict(self):
        self.rank_dict = {}

    def free(self, n, x=0):

        to_sort = list(self.rank_dict.keys())
        sort_by = list(self.rank_dict.values())
        sorter = lambda k: self.rank_dict[k]
        ranks = sorted(to_sort, key=sorter)
        del_indices = ranks[0:n]

        self.__delitem__(del_indices)

    def __delitem__(self, index):

        self.delete(index)
        # Remove index components from rank_dict
        index = index_to_list(index)
        for i in index:
            del self.rank_dict[i]
        self.on_delete(index)

    def on_update_existing(self, index, message): pass

    def on_add_new(self, index, message): pass

    def on_delete(self, index): pass

    def on_getitem(self, index): pass

    def __getitem__(self, index):

        message = super().__getitem__(index)
        self.on_getitem(index)
        return message

    def _update_existing(self, index, message):
        super()._update_existing(index, message)
        self.on_update_existing(index, message)

    def _add_new(self, index, message):
        super()._add_new(index, message)
        self.on_add_new(index, message)

class LRUCache(RankingCache, BufferedCache):
    """
    Implements a Least Recently Used cache. Items are deleted in descending order of how recently they were accessed.
    A call to __getitem__ or __setitem__ counts as accessing an element.
    """

    def __init__(self, *args, buffer_size = 0, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_rank_dict()
        self.init_buffer(buffer_size)
        self.counter = 0

    def update_rank(self, index):
        index = index_to_list(index)
        for i in index:
            self.rank_dict[i] = self.counter
        self.counter += 1

    def on_update_existing(self, index, message):
        self.update_rank(index)

    def on_add_new(self, index, message):
        self.update_rank(index)

    def on_delete(self, index): pass

    def on_getitem(self, index):
        self.update_rank(index)

class LFUCache(RankingCache, BufferedCache):
    """
    Implements a Least Frequently Used cache. Items are deleted in increasing order of how frequently they are accessed.
    A call to __getitem__ or __setitem__ counts as accessing an element.
    """
    def __init__(self, *args, buffer_size=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_rank_dict()
        self.init_buffer(buffer_size)

    def update_rank(self, index):
        index = index_to_list(index)
        for i in index:
            if i in self.pointers:
                self.rank_dict[i] += 1

    def on_update_existing(self, index, message):
        self.update_rank(index)

    def on_add_new(self, index, message):
        index = index_to_list(index)
        for i in index:
            self.rank_dict[i] = 0

    def on_delete(self, index):
        self.update_rank(index)

    def on_getitem(self, index):
        self.update_rank(index)

def pointer_adjustment_function(index):
    """
    Given an index, returns a function that that takes an integer as input and returns how many elements of the index the number is greater than.
    This is used for readjusting pointers after a deletion. For example, if you delete index 2, then every index greater than 2 must slide down 1
    but index 0 and 1 do not more.
    """
    if not type(index) is list:
        if type(index) is slice:
            index = slice_to_list(index)
        if type(index) is int:
            index = [index]

    index = sorted(index)
    def adjustment_function(x):
        for i,c in zip(index, count()):
            if x < i:
                return c
        return c+1 # If x > every index

    return adjustment_function

def index_to_list(index):
    """
    Converts an index to a list.
    """
    if type(index) is slice:
        index = slice_to_list(index)
    if type(index) is int:
        index = [index]
    return index

def slice_to_list(s):
    """
    Converts a slice object to a list of indices
    """
    step = s.step or 1
    start = s.start
    stop = s.stop
    return [x for x in range(start,stop,step)]

def get_indices(values, listlike):
    """
    Returns the indices in litlike that match elements in values
    """

    return [i for i, l  in zip(count(), listlike) if l in values]



from abc import ABC, abstractmethod
from Bio import SeqIO
import pandas as pd
from fireworks.core import Message, Junction
from fireworks.utils import index_to_list
from fireworks.core.cache import LRUCache, LFUCache, UnlimitedCache
from abc import ABC, abstractmethod
from itertools import count
import types
import random
from bidict import bidict
import torch

class HubJunction(Junction):
    """
    This junction takes multiple sources implementing __next__ as input and implements a new __next__ method that samples
    its input sources.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_inputs()

    def check_inputs(self):

        for name, source in self.components.items():
            if not (hasattr(source, '__next__') and hasattr(source, 'reset')):
                raise ValueError("Input sources must implement __next__ and reset.")

    def __next__(self):
        """
        Call __next__ on one of the input sources based on sampling algorithm until all sources run out (can run indefinetely if one
        or more inputs are infinite.)
        """
        # Choose which input to sample
        sample = self.sample_inputs()
        # Return value
        try:
            return self.components[sample].__next__()
        except StopIteration: # Remove sample from available_inputs list
            self._available_inputs.remove(sample)
            if not self._available_inputs: # Set of inputs is empty, because they have all finished iterating
                raise StopIteration
            else: # Recursively try again
                return self.__next__()

    def reset(self):
        for name, source in self.components.items():
            if name != '_available_inputs':
                source.reset()
        self._available_inputs = set(self.components.keys()) # Keep track of which sources have not yet run out.
        if '_available_inputs' in self._available_inputs:
            self._available_inputs.remove('_available_inputs')

    def __iter__(self):
        self.reset()
        return self

    @abstractmethod
    def sample_inputs(self):
        """
        Returns the key associated with an input source that should be stepped through next.
        """
        pass

class RandomHubJunction(HubJunction):
    """
    HubJunction that randomly chooses inputs to step through.
    """
    # TODO: Add support for weighted random sampling

    def sample_inputs(self):
        return random.sample(self._available_inputs, 1)[0]

class ClockworkHubJunction(HubJunction):
    """
    HubJunction that iterates through input sources one at a time.
    """
    # TODO: Add support for weighted iteration and setting order (ie. spend two cycles on one input and one on another)

    def reset(self):
        super().reset()
        self.cycle_dict = {i: name for i,name in zip(count(),self._available_inputs)}
        self.current_cycle = 0

    def sample_inputs(self):
        """
        Loops through inputs until finding one that is available.
        """
        # if not self.available_inputs:
        #     raise StopIteration
        while True:
            sample = self.cycle_dict[self.current_cycle]
            self.current_cycle = (self.current_cycle + 1) % len(self.cycle_dict)
            if sample in self._available_inputs:
                return sample

class SwitchJunction(Junction):
    """
    This junction has an internal switch that determines which of it's components all method calls will be routed to.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.switch = random.sample(self.components.keys())

    @property
    def route(self):
        """
        Returns the component to route method calls to based on the internal switch.
        """
        return self.components[self.switch]

    def __call__(self, *args, **kwargs):

        return self.route(*args, **kwargs)

    def __next__(self):

        return self.route.__next__()

    def __getitem__(self, index):

        return self.route[index]

    def __getattr__(self, attr):

        return getattr(self.route, attr)



import sqlalchemy
from sqlalchemy import Table, Column, Integer, MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Query
from fireworks.core import Message, cat, Pipe
import numpy as np
import pandas as pd

#TODO: Develop a way to infer schemas from Messages and Pipes to make save/load easier.

class TablePipe(Pipe):
    """
    Represents an SQLalchemy Table while having the functionality of a Pipe.
    """
    def __init__(self, table, engine, columns = None, input = None, **kwargs):

        super().__init__(input=input, **kwargs)
        self.sessionmaker = sessionmaker(bind=engine)
        self.reset_session()
        self.engine = engine
        if type(table) is str:
            table = reflect_table(table, engine)
        self.table = table # An SQLalchemy table class
        self.columns = columns or parse_columns(table)
        self.init_db()

    def init_db(self):
        """
        Initializes metadata for internal table. This ensures that the table exists in the database.
        """
        self.table.metadata.create_all(self.engine)

    def commit(self):
        """
        Commits transactions to database.
        """
        self.session.commit()

    def rollback(self):
        """
        Rollbacks transactions in current session that haven't been committed yet.
        """
        self.session.rollback()

    def reset_session(self, session=None):

        try:
            self.session.close()
        except:
            pass
        self.session = session or self.sessionmaker()

    def insert(self, batch):
        """
        Inserts the contents of batch message into the database using self.table object
        NOTE: Only the dataframe components of a message will be inserted.

        Args:
            batch (Message): A message to be inserted. The columns and types must be consistent with the database schema.
        """

        # TODO:     Enable inserting TensorMessages (QUESTION: how would this work?)
        rows = list(batch.df.T.to_dict().values())

        # NOTE:     Bulk inserts implemented using this guide:
        #           http://docs.sqlalchemy.org/en/latest/faq/performance.html#i-m-inserting-400-000-rows-with-the-orm-and-it-s-really-slow

        # QUESTION: Should this be the standard insert method, or should this be configurable?
        #           There are tradeoffs between doing bulk inserts and the default add method.
        #           See here for more details: http://docs.sqlalchemy.org/en/latest/orm/persistence_techniques.html#bulk-operations

        if hasattr(self.table, '__table__'):
            table = self.table.__table__
        elif type(self.table) is Table:
            table = self.table
        else:
            raise ValueError("table must either be an SQLalchemy table or have an __table__ attribute that is.")

        self.session.execute(
            table.insert(), # NOTE: This requires self.table to have a __table__ attribute, which is not guaranteed.
            rows
        )

    def query(self, entities=None, *args, **kwargs):
        """
        Queries the database and generates a DBPipe corresponding to the result.

        Args:
            entities: A list of column names
            args: Optional positional arguments for the SQLalchemy query function
            kwargs: Optional keyword arguments for the SQLalchemy query function

        Returns:
            dbpipe (DBPipe): A DBPipe object that can iterate through the results of the query.
        """
        if entities is None:
            query = self.session.query(*args, **kwargs)
            if query.column_descriptions == []:
                columns_and_types = None
            else:
                columns_and_types = parse_columns_and_types(self.table)
            # columns_and_types = None
        elif type(entities) is str:
            entities = getattr(self.table, entities)
            query = self.session.query(entities, *args, **kwargs)
            columns_and_types = parse_columns_and_types(query)
        elif type(entities) is list:
            entities = [getattr(self.table, entity) for entity in entities]
            query = self.session.query(*entities, *args, **kwargs)
            columns_and_types = parse_columns_and_types(query)

        return DBPipe(self.table, self.engine, query, columns_and_types = columns_and_types)

    def delete(self, column_name, values):
        to_delete = self.query().filter(column_name, 'in_', values)
        to_delete.delete()

    def update(self, filter_column, batch): # TODO: Support filters with multiple simultaneous columns
        to_update = self.query().filter(filter_column, 'in_', batch[filter_column])
        to_update.update(batch)

    def upsert(self, batch):
        """
        Performs an upsert into the database. This is equivalent to performing an update + insert (ie. if value is not present, insert it,
        otherwise update the existing value.)

        Args:
            batch (Message): The message to upsert.
        """
        pass


    def make_row(self, row):
        """
        Converts a Message or dict mapping columns to values into a table object that can be inserted into an SQLalchemy database.

        Args:
            row: row in Message or dict form to convert.

        Returns:
            table: row converted to table form.
        """

        kwargs = {key: cast(row[key][0]) for key in self.columns}

        return self.table(**kwargs)

    def make_row_dict(self, row):
        """
        Converts a 1-row Message into a dict of atomic (non-listlike) elements. This can be used for the bulk_insert_mappings method of
        an SQLalchemy session, which skips table instantiation and takes dictionaries as arguments instead.

        Args:
            row: row in Message or dict form to convert.

        Returns:
            table: row converted to table form.
        """

        kwargs = {key: cast(row[key][0]) for key in self.columns}

        return kwargs

def create_table(name, columns, primary_key = None, Base=None):
    """
    Creates a table given a dict of column names to data types. This is an easy
    way to quickly create a schema for a data pipeline.

    Args:
        columns (dict): Dict mapping column names to SQLalchemy types.
        primary_key: The column that should be the primary key of the table. If unspecified, a new auto-incrementing column called 'id'
            will be added as the primary key. SQLalchemy requires that all tables have a primary key, and this ensures that every row
            is always uniquely identifiable.
        Base : An optional argument that can be provided to specify the Base class that the new table class will inherit from. By
            default, this will be set to an instance of declarative_base from SQLalchemy.

    Returns:
        simpletable (sqlalchemy.ext.declarative.api.DeclarativeMeta): A table class specifying the schema for the database table.
    """
    if primary_key is None: # Create a default, autoincrementing primary key # NOTE: TODO: If a primary key is desired, we need to specify that
        columns.insert(0, Column('id', Integer, primary_key=True, autoincrement=True)) # Prepend to columns list

    Base = Base or declarative_base()

    class SimpleTable(Base):
        __table__ = Table(name, Base.metadata, *columns)

        def __repr__(self):
            return "Table {name} with values {values}".format(name=self.__table__.name,
            values = ", ".join(["{0}={1}".format(c.name, self.__getattribute__(c.name)) for c in columns]))

    return SimpleTable

class DBPipe(Pipe):
    """
    Pipe that can iterate through the output of a database query.
    """
    def __init__(self, table, engine, query = None, columns_and_types= None):
        """
        Args:
            table (sqlalchemy.ext.declarative.api.DeclarativeMeta): Table to perform query on. You can alternatively provide the name of the
                table as a string, and the schema will be extracted from the engine.
            engine (sqlalchemy.engine.base.Engine): Engine correspondign to the database to read from.
            query: Can optionally provide an SQLalchemy Query object. If unspecified, the DBPipe will perform a SELECT * query.
        """
        Pipe.__init__(self)
        self.engine = engine
        self.sessionmaker = sessionmaker(bind=engine)
        if query is None:
            self.reset_session()
            self.query = self.session.query()
        else:
            self.query = query # or self.session.query()
            self.reset_session(query.session)
        if type(table) is str:
            self.table = reflect_table(table, engine)
        else:
            self.table = table

        if columns_and_types is None:
            self.columns_and_types = parse_columns_and_types(self.query, ignore_id=False)
            if self.columns_and_types == {}: # Remap empty query to SELECT *
                self.columns_and_types = parse_columns_and_types(self.table, ignore_id=False)
                self.query = self.session.query(self.table)
        else:
            self.columns_and_types = columns_and_types

    def reset(self):

        self.__iter__()

    def __iter__(self):

        self.iterator = self.query.__iter__()
        return self

    def reset_session(self, session=None):

        if hasattr(self, 'session'):
            self.session.close()

        self.session = session or self.sessionmaker()

        if hasattr(self, 'query'):
            self.query.session = self.session

    def __next__(self):

        return to_message(self.iterator.__next__(), columns_and_types=self.columns_and_types)

    def __len__(self):

        return self.query.count()

    def filter(self, column_name, predicate, *args, **kwargs):
        """
        Applies an sqlalchemy filter to query.
        """
        column = getattr(self.table, column_name)
        predicate_function = getattr(column, predicate)
        query = self.query.filter(predicate_function(*args, **kwargs))
        filtered = DBPipe(self.table, self.engine, query, columns_and_types=self.columns_and_types)
        return filtered

    def all(self): #TODO: Test dis ish #TODO: Implement the other query methods
        """
        Returns the results of the query as a single Message object.
        """
        return cat([to_message(x, columns_and_types=self.columns_and_types) for x in self.query.all()])

    def delete(self):
        self.query.delete(synchronize_session=False)

    def update(self, batch):
        """
        Updates the contents of this DBPipe by replacing them with batch

        Args:
            batch: A Message
        """
        rows = list(batch.df.T.to_dict().values())
        self.session.bulk_update_mappings(self.table, rows)

def parse_columns(object, ignore_id=True):
    """
    Returns the names of columns in a table or query object

    Args:
        table (sqlalchemy.ext.declarative.api.DeclarativeMeta):
        ignore_id (bool): If True, ignore the 'id' column, which is a default primary key
            added by the create_table function.

    Returns:
        columns (list): A list of columns names in the sqlalchemy object.
    """

    return list(parse_columns_and_types(object, ignore_id).keys())

def parse_columns_and_types(object, ignore_id = True):
    """
    Returns column names and types in a object or query object as a dict

    Args:
        object: An SQLalchemy table or Query object
        ignore_id (bool): If True, ignore the 'id' column, which is a default primary key
            added by the create_table function.

    Returns:
        columns_and_types (dict): A dict mapping column names to their SQLalchemy type.
    """
    if type(object) is Query:
        columns_and_types = {col['name']: col['type'] for col in object.column_descriptions} #TODO: Test this
    elif hasattr(object, '__table__'):
        columns_and_types = {str(c.key): c.type for c in object.__table__.columns}
    elif type(object) is Table:
        columns_and_types = {str(c.key): c.type for c in object.columns}
    else:
        raise AttributeError("Could not extract column from table.")
    if ignore_id and 'id' in columns_and_types:
        del columns_and_types['id']
    return columns_and_types

def to_message(row, columns_and_types=None):
    """
    Converts a database query result produced by SQLalchemy into a Message

    Args:
        row: A row from the query.
        columns_and_types (dict): If unspecified, this will be inferred. Otherwise,
            you can specify the columns
            to parse, for example, if you only want to extract some columns.
    Returns:
        message: Message representation of input.
    """
    if columns_and_types is None:
        columns_and_types = parse_columns_and_types(row)

    row_dict = {c: [getattr(row,c)] for c in columns_and_types.keys()}

    return Message(row_dict)

def cast(value):
    """
    Converts values to basic types (ie. np.int64 to int)

    Args:
        value: The object to be cast.

    Returns:
        The cast object.

    """
    if type(value) is pd.Series:
        value = value[0]
    if type(value) is np.int64:
        value = int(value)
    if type(value) is np.float64:
        value = float(value)
    if type(value) is pd.Timestamp:
        value = value.to_pydatetime()

    return value

def reflect_table(table_name, engine):
    """
    Gets the table with the given name from the sqlalchemy engine.

    Args:
        table_name (str): Name of the table to extract.
        engine (sqlalchemy.engine.base.Engine): Engine to extract from.

    Returns:
        table (sqlalchemy.ext.declarative.api.DeclarativeMeta): The extracted table, which can be now be used to read from the database.
    """
    meta = MetaData()
    table = Table(table_name, meta, autoload=True, autoload_with=engine)
    return table



from sqlalchemy import Table, Column, Float, String, create_engine
from fireworks.extensions.database import create_table, TablePipe, DBPipe
from fireworks.toolbox.preprocessing import train_test_split
from fireworks.toolbox.pipes import LoopingPipe, CachingPipe, ShufflerPipe, BatchingPipe, TensorPipe
import os
import json
""" Modify the nonlinear regression examples to read from a database for input data and write to database for factory metrics """

# Read generated data from database
from nonlinear_regression_utils import generate_data

columns = [
    Column('x', Float),
    Column('y', Float),
    Column('errors', Float),
]

table = create_table("nonlinear_regression", columns)

def write_data(filename='example.sqlite', n=1000):
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass

    engine = create_engine('sqlite:///{0}'.format(filename))
    db = TablePipe(table, engine)

    data, params = generate_data(n)
    db.insert(data)

    with open(filename+"_params", 'w') as f:
        f.write(json.dumps(params))

    db.commit()

def load_data(filename='example.sqlite'):
    if not os.path.exists(filename):
        raise FileNotFoundError("File {0} does not exist.".format(filename))
    with open(filename+'_params') as f:
        params = json.load(f)

    engine = create_engine('sqlite:///{0}'.format(filename))
    db = DBPipe(table, engine) # Default query is SELECT * FROM table

    return db, params

def get_data(filename='example.sqlite', n=1000):
    if not os.path.exists(filename) and os.path.exists(filename+'_params'):
        write_data(filename, n)

    data, params = load_data(filename)
    looper = LoopingPipe(data)
    cache = CachingPipe(looper, cache_size=1000)
    train, test = train_test_split(cache, test=.25)

    shuffler = ShufflerPipe(train)
    minibatcher = BatchingPipe(shuffler, batch_size=25)
    train_set = TensorPipe(minibatcher, columns=['x','y'])

    test_set = TensorPipe(test, columns=['x','y'])

    return train_set, test_set, params



from fireworks.toolbox import text
import random

def test_space_tokenizer():
    phrase = "Hello is this dog"
    tokens = text.space_tokenizer(phrase)
    assert (tokens == ['Hello', 'is', 'this', 'dog'])

def test_character_tokenizer():

    sequence = 'hello'
    split = text.character_tokenizer(sequence)
    assert split == ['h', 'e', 'l', 'l', 'o']

test_embedding_dict = {str(i):i for i in range(80)}
test_embedding_dict['EOS'] = 81
test_embedding_dict['SOS'] = 82
test_embedding_dict['UNK'] = 83

def test_pad_sequence():

    width = 100
    test_sequence = lambda : [random.randint(0,90) for _ in range(random.randint(1,width))]
    sequence = test_sequence()
    n = len(sequence)
    pad = 150
    padded = text.pad_sequence(sequence, pad, embeddings_dict = test_embedding_dict)
    assert (padded[0:n] == sequence).all()
    for p in padded[n:pad]:
        assert p == test_embedding_dict['EOS']
    # Test with a different pad_token
    padded = text.pad_sequence(sequence, pad, embeddings_dict = test_embedding_dict, pad_token='2')
    assert (padded[0:n] == sequence).all()
    for p in padded[n:pad]:
        assert p == test_embedding_dict['2']

    # count = 10
    # sequences = [test_sequence() for i in range(count)]
    # padded = [text.pad_sequence(sequence, n, embeddings_dict = test_embedding_dict) for sequence in sequences]
    # assert len(padded) == count
    # assert padded.shape == (count, width)

def test_pad():

    lengths = [random.randint(10, 15) for _ in range(10)]
    test_sequence = lambda n: [1 for _ in range(n)]
    sequences = [test_sequence(i) for i in lengths]
    padded = text.pad(sequences, test_embedding_dict)
    max_length = max(lengths)
    for original, new in zip(sequences, padded):
        assert len(new) == max_length
        a = len(original)
        b = max_length
        for token in new[a:b]:
            assert token == test_embedding_dict['EOS']

def test_apply_embeddings():

    phrase = '1234x'
    embeddings = text.apply_embeddings(phrase, embeddings_dict=test_embedding_dict, tokenizer = text.character_tokenizer)
    assert len(embeddings) == 7
    assert embeddings[0] == test_embedding_dict['SOS']
    assert embeddings[1] == test_embedding_dict['1']
    assert embeddings[2] == test_embedding_dict['2']
    assert embeddings[3] == test_embedding_dict['3']
    assert embeddings[4] == test_embedding_dict['4']
    assert embeddings[5] == test_embedding_dict['UNK']
    assert embeddings[6] == test_embedding_dict['EOS']


# def test_create_pretrained_embeddings():
#     embeddings = nlp.create_pretrained_embeddings(os.path.join(fireworks.EXTERNAL_DIR, 'glove.test.txt'))
#     assert len(embeddings.keys()) == 18
#     for key, item in embeddings.items():
#         assert len(item) == 300

def test_too_big() : pass



from fireworks import message as messi
from fireworks import Message, TensorMessage
import torch
import os
import numpy as np
from itertools import product
import pandas as pd
from itertools import count

tensors = {
    'a': torch.Tensor([1,2,3]),
    'b': torch.Tensor([4,5,6]),
    }
vectors = {
    'c': np.array([7,8,9]),
    'd': np.array([10,11,12]),
}
dtensors = {
    'a': torch.Tensor([[1,2,3],[4,5,6],[7,8,9]]),
    'b': torch.Tensor([[-1,-2,-3],[-4,-5,-6], [-7,-8,-9]]),
}

def test_compute_length():

    l = messi.compute_length(tensors)
    assert l == 3
    l = messi.compute_length(vectors)
    assert l == 3

def test_extract_tensors():

    target = {**tensors, **vectors}
    t, v = messi.extract_tensors(target)
    assert t == tensors
    assert v == vectors
    t, v = messi.extract_tensors(tensors)
    assert t == tensors
    assert v == {}
    t, v = messi.extract_tensors(vectors)
    assert t == {}
    assert v == vectors

def test_complement():

    n = 10
    index = 7
    complement = messi.complement(index, n)
    assert complement == [0,1,2,3,4,5,6,8,9]
    index = slice(2,5)
    complement = messi.complement(index, n)
    assert complement == [0,1,5,6,7,8,9]
    index = [2,4,6]
    complement = messi.complement(index, n)
    assert complement == [0,1,3,5,7,8,9]

def test_Message():
    """ Test init, getitem, and len methopl. """

    def attribute_test(message, length = 3):
        assert len(message) == length
        assert message[0].tensors() == {
            'a': torch.Tensor([1]),
            'b': torch.Tensor([4]),
        }
        assert message[0].dataframe().equals(pd.DataFrame({
            'c': np.array([7]),
            'd': np.array([10]),
        }))

        assert message[0] == Message({'a': torch.Tensor([1]),'b': torch.Tensor([4])}, pd.DataFrame({'c': np.array([7]),'d': np.array([10]),}))

        assert message[1:3].tensors() == {
            'a': torch.Tensor([2,3]),
            'b': torch.Tensor([5,6]),
        }
        assert message[1:3].dataframe().equals(pd.DataFrame({
            'c': np.array([8,9]),
            'd': np.array([11,12]),
        }))

        assert (message['a'] == torch.Tensor([1,2,3])).all()
        assert message[['a','c']] == Message({'a': torch.Tensor([1,2,3]), 'c': np.array([7,8,9])})

        assert message[1:3] == Message({'a': torch.Tensor([2,3]),'b': torch.Tensor([5,6])}, pd.DataFrame({'c': np.array([8,9]),'d': np.array([11,12])}))

        # Test length
        assert len(message) == length
        # Test __getitem__

    # Init empty message
    m = Message()
    assert len(m) == 0
    # Init message from tensor_dict / TensorMessage and dict of arrays / dataframe using positional arguments.
    tensor_message = TensorMessage(tensors)
    tensor_as_message = Message(tensors = tensors)
    df = pd.DataFrame(vectors)
    df_as_message = Message(df = vectors)

    # Try every combination
    tensor_options = [tensors, tensor_message, tensor_as_message]
    vector_options = [vectors, df, df_as_message]

    for t, v in product(tensor_options, vector_options):
        m = Message(t, v)
        attribute_test(m)
        m = Message(tensors = t, df = v)
        attribute_test(m)

    # Test one sided Messages
    for t in tensor_options:
        m = Message(t, None)
        assert len(m) == 3
        assert m == Message(tensors)
    for v in vector_options:
        m = Message(None, v)
        assert len(m) == 3
        assert m == Message(vectors)

    # Init message from a single dict
    everything = {**tensors, **vectors}
    m = Message(everything)
    attribute_test(m)

def test_Message_from_objects():

    v = vectors.copy()
    t = tensors.copy()
    v['c'] = np.array([1.,2.])
    v['r'] = 'howdy'
    t['a'] = torch.randn(5)
    t['q'] = torch.randn([4,3])
    combined = {**t, **v}

    m = Message.from_objects(t, v)
    assert (set(m.keys()) == set(['c','d','r','b','a','q']))
    for key in ['c','d','b','a','q']:
        assert (m[key][0] == combined[key]).all()
    assert m['r'][0] == combined['r']
    assert len(m) == 1

def test_getitem():

    m = Message(tensors, vectors)
    assert m[0] == Message({'a': torch.Tensor([1]), 'b': torch.Tensor([4])}, {'c': np.array([7]), 'd': np.array([10])})
    assert m[[0,2]] == Message({'a': torch.Tensor([1,3]), 'b': torch.Tensor([4,6])}, {'c': np.array([7,9]), 'd': np.array([10,12])})
    # Check that out of bounds index calls raise errors
    try:
        m[3]
        assert False
    except IndexError:
        assert True

    try:
        m[3:5]
        assert False
    except IndexError:
        assert True

def test_cache(): pass

def test_tensors():
    m = Message(tensors, vectors)
    t = m.tensors()
    assert t == TensorMessage(tensors)
    t = m.tensors(keys=['a'])
    assert t == TensorMessage({'a': tensors['a']})
    t = m.tensors(keys=['a','c'])
    assert t == TensorMessage({'a': tensors['a'], 'c': torch.Tensor(vectors['c'])})

def test_df():
    m = Message(tensors, vectors)
    df = m.dataframe()
    assert df.equals(pd.DataFrame(vectors))
    df = m.dataframe(keys=['c'])
    assert df.equals(pd.DataFrame({'c': vectors['c']}))
    df = m.dataframe(keys=['c','a'])
    assert (df == (pd.DataFrame({'c': vectors['c'], 'a': np.array(tensors['a'])}))).all().all()

def test_to_dataframe():

    mo = Message(tensors,vectors)
    # no = mo.to_dataframe()
    # assert no.tensor_message == {}
    # assert (no['a'] == mo['a']).all()
    # assert (no['b'] == mo['b']).all()
    # for letter in ['a','b','c','d']:
    #     assert letter in no.df
    lo = Message(dtensors, vectors)
    ok = lo.to_dataframe()
    for i in range(3):
        assert (ok['a'][i] == dtensors['a'][i].numpy()).all()
        assert (ok['b'][i] == dtensors['b'][i].numpy()).all()

def test_cpu_gpu():
    m = Message(tensors, vectors)
    m.cpu()
    assert set(m.tensors().keys()) == set(['a','b'])
    for key, tensor in m.tensors().items():
        assert tensor.device.type == 'cpu'
    if torch.cuda.is_available():
        m.cuda()
        for key, tensor in m.tensors().items():
            assert tensor.device.type == 'cuda'
        m.cpu()
        for key, tensor in m.tensors().items():
            assert tensor.device.type == 'cpu'

def test_append():

    t = tensors
    v = vectors
    m1 = Message(t, v)
    m2 = Message(t, v)
    m3 = Message(t)
    m4 = TensorMessage(t)
    m5 = Message(pd.DataFrame(v))
    m6 = pd.DataFrame(v)

    m0 = Message()
    assert(len(m0) == 0)
    m = m0.append(Message(t))
    assert m == Message(t)
    m = m0.append(Message(v))
    assert m == Message(v)
    m = m0.append(Message(t,v))
    assert m == Message(t,v)

    m = m1.append(m2)
    assert len(m) == 6
    assert m == Message({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])}, {'c': np.array([7,8,9,7,8,9]), 'd': np.array([10,11,12,10,11,12])})
    m = m3.append(t)
    assert len(m) == 6
    assert m == Message({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})
    m = m3.append(m3)
    assert len(m) == 6
    assert m == Message({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})
    m = m3.append(m4)
    assert len(m) == 6
    assert m == Message({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})

    m = m4.append(t)
    assert len(m) == 6
    assert m == TensorMessage({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})
    m = m4.append(m3)
    assert len(m) == 6
    assert m == TensorMessage({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})
    m = m4.append(m4)
    assert len(m) == 6
    assert m == TensorMessage({'a': torch.Tensor([1,2,3,1,2,3]), 'b': torch.Tensor([4,5,6,4,5,6])})

    m = m5.append(v)
    assert len(m) == 6
    assert m == Message({'c': np.array([7,8,9,7,8,9]), 'd': np.array([10,11,12,10,11,12])})
    m = m5.append(m5)
    assert len(m) == 6
    assert m == Message({'c': np.array([7,8,9,7,8,9]), 'd': np.array([10,11,12,10,11,12])})
    m = m5.append(m6)
    assert len(m) == 6
    assert m == Message({'c': np.array([7,8,9,7,8,9]), 'd': np.array([10,11,12,10,11,12])})

    # Test type conversions on appending to TensorMessage
    m = m4.append({'a': np.array([42]), 'b': np.array([24])})
    assert len(m) == 4
    assert m == TensorMessage({'a': torch.Tensor([1,2,3,42]), 'b': torch.Tensor([4,5,6,24])})


def test_join():

    t = tensors
    v = vectors
    t2 = {'d': torch.Tensor([13,14,15])}
    v2 = {'e': np.array([16,17,18])}
    m1 = Message(t,v)
    m2 = Message(t)
    m2_t = TensorMessage(t)
    m3 = Message(v)
    m4 = Message(t2,v2)
    m5 = Message(t2)
    m5_t = TensorMessage(t2)
    m6 = Message(v2)
    m7 = Message(t,v2)
    m8 = Message(t2, v)

    # Test if a tensor message can be merged into a message and vice versa
    assert m2.merge(m3) == m1
    assert m3.merge(m2) == m1
    assert m3.merge(m2_t) == m1
    assert m3.merge(t) == m1

    # Test if the tensors in messages can be merged
    assert m2.merge(t2) == Message({**t, **t2})
    assert m2.merge(m5) == Message({**t, **t2})
    assert m2.merge(m5_t) == Message({**t, **t2})
    assert m2_t.merge(t2) == TensorMessage({**t, **t2})
    assert m2_t.merge(m5) == TensorMessage({**t, **t2})
    assert m2_t.merge(m5_t) == TensorMessage({**t, **t2})

    # Test if the dataframes in messages can be merged
    assert m3.merge(m6) == Message({**v, **v2})
    assert m6.merge(m3) == Message({**v, **v2})
    assert m3.merge(v2) == Message({**v, **v2})

def test_Message_set_get():

    # Test point updates
    email = Message(tensors, vectors)
    gmail = Message({'a':torch.Tensor([1,42,3]), 'b':torch.Tensor([4,43,6]), 'c': np.array([7,99,9]), 'd': np.array([10,100,12])})
    replacement = {'a': torch.Tensor([42]), 'b': torch.Tensor([43]), 'c': np.array([99]), 'd': np.array([100])}
    assert len(email) == 3
    assert email != gmail
    email[1] = replacement
    assert email == gmail
    # Test ranged updates
    email = Message(tensors, vectors)
    gmail = Message({'a':torch.Tensor([1,42,33]), 'b':torch.Tensor([4,43,66]), 'c': np.array([7,99,99]), 'd': np.array([10,100,122])})
    replacement = {'a': torch.Tensor([42,33]), 'b': torch.Tensor([43,66]), 'c': np.array([99,99]), 'd': np.array([100,122])}
    assert email != gmail
    email[1:3] = replacement
    assert email == gmail
    # Test updates using lists as indexes
    email = Message(tensors, vectors)
    assert email != gmail
    email[[1,2]] = replacement
    assert email == gmail
    # Test column updates
    email['a'] = torch.Tensor([9,9,9])
    assert torch.equal(email['a'], torch.Tensor([9,9,9]))
    email['c'] = np.array([9,9,9])
    assert email['c'].equals(pd.Series([9,9,9]))
    # Test column updates that switch from df to tensor and vice-versa
    email = Message(tensors, vectors)
    assert set(email.columns) == set(['a','b','c','d'])
    assert set(email.tensor_message.columns) == set(['a','b'])
    assert set(email.df.columns) == set(['c','d'])
    new_a = np.array([1,2,3]) # Switch from tensor to vector
    email['a'] = new_a
    assert set(email.columns) == set(['a','b','c','d'])
    assert set(email.tensor_message.columns) == set(['b'])
    assert set(email.df.columns) == set(['a','c','d'])
    assert (email['a'] == new_a).all()
    new_c = torch.Tensor([7,8,9])
    email['c'] = new_c
    assert set(email.columns) == set(['a','b','c','d'])
    assert set(email.tensor_message.columns) == set(['b','c'])
    assert set(email.df.columns) == set(['a','d'])
    assert (email['c'] == new_c).all()
    # Test column updates that end up clearing either self.df or self.tensor_message
    email = Message(tensors, vectors)
    df = email.dataframe(['a', 'b'])
    assert len(email) == 3
    assert len(email.tensor_message) == 3
    assert len(email.df) == 3
    email[['a','b']] = df
    assert len(email) == 3
    assert len(email.tensor_message) == 0
    assert len(email.df) == 3
    # TODO: Test the other way around

def test_Message_del():

    t = {
        'a': torch.Tensor([1,2,3]),
        'b': torch.Tensor([4,5,6]),
        }
    v = {
        'c': np.array([7,8,9]),
        'd': np.array([10,11,12]),
    }
    t2 = {
        'a': torch.Tensor([1,2]),
        'b': torch.Tensor([4,5]),
        }
    v2 = {
        'c': np.array([7,8]),
        'd': np.array([10,11]),
    }
    t3 = {
        'a': torch.Tensor([1]),
        'b': torch.Tensor([4]),
        }
    v3 = {
        'c': np.array([7]),
        'd': np.array([10]),
    }

    # Test deletions for messages with only tensors, only df, and both
    # Test point deletions
    m = Message(t,v)
    m1 = Message(t)
    m2 = Message(v)
    assert m != Message(t2,v2)
    assert m1 != Message(t2)
    assert m2 != Message(v2)
    assert len(m) == 3
    assert len(m1) == 3
    assert len(m2) == 3
    del m[2]
    del m1[2]
    del m2[2]
    assert len(m) == 2
    assert len(m1) == 2
    assert len(m2) == 2
    assert m == Message(t2,v2)
    assert m1 == Message(t2)
    assert m2 == Message(v2)
    # Test range deletions
    m = Message(t,v)
    m1 = Message(t)
    m2 = Message(v)
    assert m != Message(t3,v3)
    assert m1 != Message(t3)
    assert m2 != Message(v3)
    assert len(m) == 3
    assert len(m1) == 3
    assert len(m2) == 3
    del m[1:3]
    del m1[1:3]
    del m2[1:3]
    assert len(m) == 1
    assert len(m1) == 1
    assert len(m2) == 1
    assert m == Message(t3,v3)
    assert m1 == Message(t3)
    assert m2 == Message(v3)
    # Test list deletions
    m = Message(t,v)
    m1 = Message(t)
    m2 = Message(v)
    assert m != Message(t3,v3)
    assert m1 != Message(t3)
    assert m2 != Message(v3)
    assert len(m) == 3
    assert len(m1) == 3
    assert len(m2) == 3
    del m[[1,2]]
    del m1[[1,2]]
    del m2[[1,2]]
    assert len(m) == 1
    assert len(m1) == 1
    assert len(m2) == 1
    assert m == Message(t3,v3)
    assert m1 == Message(t3)
    assert m2 == Message(v3)

    # Test column deletions
    m = Message(t,v)
    assert set(m.columns) == set(['a','b','c','d'])
    del m['a']
    assert set(m.columns)== set(['b','c','d'])
    del m['c']
    assert set(m.columns) == set(['b','d'])

def test_Message_iter():

    m = Message(tensors, vectors)
    l = len(m)
    for x,i in zip(m, count()):
        assert type(x) is Message
        if i > l:
            assert False
    assert i == l - 1

    t = TensorMessage(tensors)
    l = len(t)
    for x,i in zip(t, count()):
        assert type(x) is TensorMessage
        if i > l:
            assert False
    assert i == l - 1

def test_map(): pass

def test_TensorMessage():

    a = [1,2,3]
    b = [4, 5, 6]

    # Test init
    empty = TensorMessage()
    assert len(empty) == 0
    assert empty.keys() == {}.keys()
    email = TensorMessage({'a': a, 'b':b})

    #   Test error cases
    # TODO: test error cases

    # Test length
    assert len(email) == 3
    # Test getitem
    x = email[2]
    assert set(x.keys()) == set(['a','b'])
    assert (x['a'] == torch.Tensor([3])).all()
    assert (x['b'] == torch.Tensor([6])).all()
    x = email[0:2]
    assert set(x.keys()) == set(['a','b'])
    assert (x['a'] == torch.Tensor([1,2])).all()
    assert (x['b'] == torch.Tensor([4,5])).all()

    # Test for length 1 init
    gmail = TensorMessage({'a':1, 'b': 80})
    assert len(gmail) == 1
    y = gmail[0]
    assert set(y.keys()) == set(['a','b'])
    assert (y['a'] == torch.Tensor([1])).all()
    assert (y['b'] == torch.Tensor([80])).all()

    # Test extend
    yahoomail = email.append(gmail)
    assert len(yahoomail) == 4
    z = yahoomail[0:4]
    assert set(z.keys()) == set(['a','b'])
    assert (z['a'] == torch.Tensor([1,2,3,1])).all()
    assert (z['b'] == torch.Tensor([4, 5, 6, 80])).all()

def test_TensorMessage_set_get_del():

    a = [1,2,3]
    b = [4, 5, 6]
    email = TensorMessage({'a': a, 'b':b})
    replacement = {'a': torch.Tensor([42]), 'b': torch.Tensor([43])}
    gmail = TensorMessage({'a':torch.Tensor([42,2,3]), 'b':  torch.Tensor([43,5,6])})
    yahoomail = TensorMessage({'a':torch.Tensor([2,3]), 'b': torch.Tensor([5,6])})
    assert email != gmail
    email[0] = replacement
    assert email == gmail
    assert len(email) == 3
    email['a'] = torch.Tensor([9,9,9])
    assert torch.equal(email['a'], torch.Tensor([9,9,9]))
    assert gmail != yahoomail
    del gmail[0]
    assert len(gmail) == 2
    assert gmail == yahoomail
    # Test column deletions
    email = TensorMessage({'a': a, 'b':b})
    assert set(email.columns) == set(['a','b'])
    del email['a']
    assert set(email.columns) == set(['b'])
    # Test that out of bounds requests raise errors
    try:
        email[3]
        assert False
    except IndexError:
        assert True
    try:
        email[3:5]
        assert False
    except IndexError:
        assert True

    # Test length adjustment if all columns are deleted
    zohomail = TensorMessage({'a':a,'b':b})
    assert len(zohomail) == 3
    del zohomail['a']
    assert len(zohomail) == 3
    del zohomail['b']
    assert len(zohomail) == 0

def test_TensorMessage_eq():

    a = [1,2,3]
    b = [4, 5, 6]

    # Test init
    email = TensorMessage({'a': a, 'b':b})
    gmail = TensorMessage(email)

def test_cat():

    m = Message(tensors, vectors)
    m0 = m[0]
    m1 = m[1]
    m2 = m[2]
    babaghanush = messi.cat([m0,m1,m2])
    assert babaghanush == m

def test_TensorMessage_permute():
    a = [1,2,3]
    b = [4, 5, 6]
    email = TensorMessage({'a': a, 'b':b})
    gmail = email.permute([2,1,0])
    assert gmail == TensorMessage({'a':[3,2,1], 'b':[6,5,4]})
    gmail = email.permute([0,0,0])
    assert gmail == TensorMessage({'a':[1,1,1], 'b':[4,4,4]})

def test_permute():

    tensors = {
        'a': torch.Tensor([1,2,3]),
        'b': torch.Tensor([4,5,6]),
        }
    vectors = {
        'c': np.array([7,8,9]),
        'd': np.array([10,11,12]),
    }
    email = Message(tensors, vectors)
    gmail = email.permute([2,1,0])
    assert gmail == Message({'a':[3,2,1], 'b':[6,5,4]}, {'c': np.array([9,8,7]), 'd': np.array([12,11,10])})
    gmail = email.permute([0,0,0])
    assert gmail == Message({'a':[1,1,1], 'b':[4,4,4]}, {'c': np.array([7,7,7]), 'd': np.array([10,10,10])})
    # Test with only tensors
    email = Message(tensors)
    gmail = email.permute([2,1,0])
    assert gmail == Message({'a':[3,2,1], 'b':[6,5,4]}, {})
    gmail = email.permute([0,0,0])
    assert gmail == Message({'a':[1,1,1], 'b':[4,4,4]}, {})
    # Test with only dataframes
    email = Message(vectors)
    gmail = email.permute([2,1,0])
    assert gmail == Message({'c': np.array([9,8,7]), 'd': np.array([12,11,10])})
    gmail = email.permute([0,0,0])
    assert gmail == Message({'c': np.array([7,7,7]), 'd': np.array([10,10,10])})

def test_to_csv():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_to_pickle():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_to_sql():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_to_dict():
    m = Message(tensors, vectors)
    md = m.to_dict()
    assert type(md) is dict
    assert (md['c'] == md['c'])
    assert (md['d'] == md['d'])
    assert (md['a'] == np.array(md['a'])).all()
    assert (md['b'] == np.array(md['b'])).all()

def test_to_excel():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_to_json():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_to_string():
    m = Message(tensors, vectors)
    pass #TODO: Implement

def test_save_load():
    m = Message(tensors, vectors)
    test_path = 'test.fireworks'
    m.save(test_path)
    new_m = Message.load(test_path)
    assert new_m == m
    os.remove(test_path)



import collections
import torch
import pandas as pd
import numpy as np
import os 
import io
import pyarrow
import tarfile
from copy import deepcopy
from collections import Hashable
from fireworks.utils import index_to_list, slice_length

to_methods = ['csv', 'json', 'dict', 'html', 'feather', 'latex', 'stata', 'msgpack', 'gbq', 'records', 'sparse', 'dense', 'string', 'clipboard']
read_methods = {
    'json': pd.read_json, 'csv': pd.read_csv, 'excel': pd.read_excel, 'hdf': pd.read_hdf, 'parquet': pd.read_parquet,
    'pickle': pd.read_pickle, 'sql_table': pd.read_sql_table, 'state': pd.read_stata, 'table': pd.read_table,
    }

class Message:
    """
    A Message is a class for representing data in a way that can be consumed by different analysis pipelines in python. It does this by
    representing the data as a dictionary of arrays. This is the same approach that Pandas takes, but Messages are specifically designed to
    be used with PyTorch, in which the core data structure is a tensor, which cannot be mixed into a Pandas dataframe.
    Hence, a Message has two elements: a TensorMessage which specifically stores tensor objects and a dataframe which can be used for handling
    any other type of array data. With this structure, you could put your training data in as a TensorMessage and any associated metadata as
    the df and lump all of that into a message. All of the existing df methods can be run on the metadata, and any PyTorch operations can be
    performed on the tensor part.

    Messages support operations such as appending one message to another, joining two messages along keys, applying maps to the message's values,
    and equality checks.

    Additionally, messages behave as much as possible like dicts. In many scenarios, a dict will be able to substitute for a message and vice-versa.
    For example, for preparing batches to feed into an RNN to classify DNA sequences, one could create a Message like this:

    ::

        my_message = Message({
        'embedded_sequences': torch.Tensor([...]),
        'labels': torch.LongTensor([...]),
        'raw_sequences': ['TCGA...',...,],
        ...})

    The Message constructor will parse this dictionary and store the labels and embedded sequences inside a TensorMessage and the
    raw_sequences and other metadata in the dataframe.

    Now we can access elements of this Message:

    ::

        my_message[0]
        my_message[2:10]
        my_message['labels']
        len(my_message)

    We can also move tensors to the GPU and back:

    ::

        my_message.cpu()
        my_message.cuda(device_num=1) # Defaults to 0 for device number
        my_message.cuda(keys=['labels']) # Can specify only certain columns to move if desired

    """

    def __init__(self, *args, metadata = None, length = None, **kwargs):
        """
        Initializes a message
        If *args is empty --> empty message
        If *args is two args --> construct a message assuming arg0 is tensors and arg1 is dataframe
        Alternatively, tensors and dataframes can be specified with kwargs (tensors  = ... and df = ...)
        If *args has __getitem__ and keys() attributes --> construct a message from elements of dict, separating tensors from pd series
        """

        """ The initializer is designed to be idempotent, perform type conversions as necessary, and ensure consistency of lengths. """

        if len(args) == 0:
            tensors = {}
            df = {}

        if len(args) == 2:
            tensors = args[0]
            df = args[1]

        if len(args) == 1:
            # Identify tensors and separate them out
            # The argument may be an existing Message/TensorMessage
            if type(args[0]) is Message:
                tensors = args[0].tensor_message
                df = args[0].df

            elif args[0] is None:
                tensors = {}
                df = {}

            else:
                tensors, df = extract_tensors(args[0])

        if 'tensors' in kwargs:
            tensors = kwargs['tensors']
        if 'df' in kwargs:
            df = kwargs['df']

        self.tensor_message = TensorMessage(tensors) # The TensorMessage constructor will handle type conversions as needed.

        if type(df) is Message:
            df = df.dataframe() # Pull out the df from the message
        self.df = pd.DataFrame(df)
        # Ensure columns are in the same order always
        self.df = self.df.reindex(sorted(self.df.columns), axis=1)

        if length is None:
            self.check_length()
        else:
            self.length = length

        self._current_index = 0 # For iteration

    @classmethod
    def from_objects(cls, *args, **kwargs):
        """
        Returns a message by treating all of the provided values as atomic elments and ignoring length differences. This results in a Message
        of length 1. This can be useful if you want to treat an entire blob as data as a single unit. For example, if you want a row of the
        Message to contain the state of a neural network, one column could be 'bias', and the element could be an entire bias-vector, and
        another column could be 'weights', containing an entire weight matrix in a single row.
        """

        if len(args) == 0:
            tensors = {}
            df = {}

        if len(args) == 2:

            tensors = {key: value.reshape(1,*value.shape) for key, value in args[0].items()}
            df = {key: [value] for key, value in args[1].items()}

        if len(args) == 1:
            if type(args[0]) is Message:
                tensors = {key: value.reshape(1,*value.shape) for key, value in args[0].tensor_message.items()}
                df = {key: [value] for key, value in args[1].df.items()}
            else:
                tensors = {}
                df = {}
                for key, value in args[0].items():
                    if type(value) is torch.Tensor or type(value) is torch.nn.Parameter:
                        tensors[key] = value.reshape(1,*value.shape)
                    else:
                        df[key] = [value]

        if 'tensors' in kwargs:
            tensors = {key: value.reshape(1,*value.shape) for key, value in kwargs['tensors'].items()}
        if 'df' in kwargs:
            df = {key: [value] for key, value in kwargs['df'].items()}

        return cls(tensors, df)

    @classmethod
    def read(cls, method, path, *args, **kwargs):
        """
        Reads the file located at a provided path using a given method and loads the data into a Message. This method uses the Pandas .read_
        functions and converts the resulting DataFrame to a Message. Thus, there is no way currently to go straight from data in a file to
        a TensorMessage. Additionally, only the methods in the read_methods dict at the top of the file are supported.

        You can provide additional positional and key-word arguments which will be passed to the relevant Pandas method to parameterize the
        read.

        Args:
            - method: The method to read the file as. Must be one of: json, csv, excel, hdf, parquet, pickle, sql_table, stata, table.
            - path: The location of the file. This file must exist, be readable, and must be the type specified by method.
            - *args, **kwargs: Additional arguments for the underlying Pandas function call.

        Returns:
            - message: A Message representation of the loaded data.
        """
        if method in read_methods:
            df = read_methods[method](path, *args, **kwargs)
            return cls(df)
        else:
            raise TypeError("Method {0} is not a supported file type for reading. Must be one of {1}".format(method, read_methods.keys()))

    def to(self, method, *args, **kwargs):
        """
        Converts all elements of Message to dataframe and then calls df.to_x based on the given method
        This can be used to serialize and save Messages or convert them into a different format

        Args:
            - method: The method to save the file as. Must be one of: 'json', 'dict', 'html', 'feather', 'latex', 'stata', 'msgpack', 'gbq', 'records', 'sparse', 'dense', 'string', 'clipboard'
            - *args, **kwargs: Additional arguments for the underlying Pandas pd.to_ function call as desired. If the first argument (or kwarg
                               path_or_buf) is provided, it will be used as a filepath to save to.

        Returns:
            - Depending on the method chosen, this will save the Message to a file location and then return the converted Message (eg. to_json will return a json)
        """
        # The conversion to dataframe is essential, because that enforces that all of the data being serialized is on the CPU.
        if method in to_methods:
            tensor_message = deepcopy(self.tensor_message) # This is copied so that the original message is not perturbed by serialization.
            to_serialize = Message(tensor_message, self.df)
            to_serialize = to_serialize.to_dataframe().df # Convert everything to dataframe
            if 'path' in kwargs:
                kwargs['path_or_buf'] = kwargs['path']
                del kwargs['path']
            return getattr(to_serialize, 'to_{0}'.format(method))(*args, **kwargs)
        else:
            raise ValueError("Method {0} is not a valid method for Message serialization. Valid methods include {1}.".format(method, to_methods))

    def to_csv(self, *args, **kwargs):
        """ Converts Message to a CSV. """
        return self.to('csv', *args, **kwargs)

    def to_pickle(self, *args, **kwargs):
        """ Converts Message to a Python Pickle. """
        return self.to('pickle', *args, **kwargs)

    def to_sql(self, *args, **kwargs):
        """ Saves Message to a SQL Database using pd.to_sql function. """
        return self.to('sql', *args, **kwargs)

    def to_dict(self, *args, **kwargs):
        """ Converts Message to a dict. """
        if 'orient' not in kwargs:
            kwargs['orient']  = 'list'
        return self.to('dict', *args, **kwargs)

    def to_excel(self, *args, **kwargs):
        """ Saves Message to an Excel file. """
        return self.to('excel', *args, **kwargs)

    def to_json(self, *args, **kwargs):
        """ Converts Message to a JSON. """
        if 'orient' not in kwargs:
            kwargs['orient']  = 'list'
        return self.to('json', *args, **kwargs)

    def to_string(self, *args, **kwargs):
        """ Converts Message to a string. """
        return self.to('string', *args, **kwargs)

    def save(self, path):
        """
        Converts message to a custom serialized format which retains column structure for tensor and DataFrame components.
        The format is essentially a tarfile containing:
        - Parquet representation of the DataFrame
        - Pickle representation of each tensor
        - JSON mapping column names to their values
        """

        with tarfile.open(path, "w:gz") as tar:
            # Save DataFrame
            df_buffer = io.BytesIO()
            self.df.to_parquet(df_buffer)
            df_buffer.seek(0)
            df_info = tarfile.TarInfo("df.parquet")
            df_info.size = len(df_buffer.read())
            df_buffer.seek(0)
            tar.addfile(df_info, df_buffer)
            df_buffer.close()

            # Save tensors one by one
            for key, value in self.tensor_message.items():
                buffer = io.BytesIO()
                torch.save(value, buffer)
                buffer.seek(0) # Make sure we are at beginning of buffer after writing
                info = tarfile.TarInfo("{key}.torch".format(key=key))
                info.size = len(buffer.read())
                buffer.seek(0)
                tar.addfile(info, buffer)
                buffer.close()

    @classmethod
    def load(cls, path):
        """
        Loads in serialized message at the given path.        
        """
        
        df = None
        tensors = {}
        with tarfile.open(path, "r:gz") as tar:
            for filename in tar.getnames():
                if filename.endswith('.parquet'): # Is the DataFrame
                    buffer = tar.extractfile(tar.getmember(filename))
                    df = pd.read_parquet(buffer)
                elif filename.endswith('.torch'): # Is a Tensor
                    buffer = tar.extractfile(tar.getmember(filename))
                    name = filename.rstrip('.torch')
                    tensors[name] = torch.load(buffer)
        return cls(tensors, df)


    def check_length(self):
        """
        Checks that lengths of the internal tensor_message and dataframe are the same and equalto self.len
        If one of the two is empty (length 0), then that is fine.
        """
        if len(self.df) == 0:
            self.length = len(self.tensor_message)
        elif len(self.tensor_message) == 0:
            self.length = len(self.df)
        elif len(self.tensor_message) != len(self.df):
            raise ValueError("Every element of the message, including tensors and arrays, must have the same length unless one or both are None.")
        else:
            self.length = len(self.tensor_message)

    def __iter__(self):

        self._current_index = 0
        return self

    def reset(self):

        return self.__iter__()
    
    def __next__(self):

        try:
            iteration = self[self._current_index]
            self._current_index += 1
            return iteration
        except (KeyError, IndexError) as e:
            raise StopIteration

    def __len__(self):
        """
        The length of a Message is the length of its internal DataFrame and/or internal TensorFrame. These two must have the same length
        for consistency. In this way, row i of a Message is row i of the DataFrame and row i of the TensorFrame.
        The only exception to this rule is if one of the two are empty. In that case, the empty component is ignored. However, the moment you
        insert data into the empty component (eg. insert a Tensor into a Message that previously only had DataFrame elements), then the lengths
        must again be equal (ie. you can only insert a Tensor that has the same length as the Message.)
        """
        return self.length

    def __eq__(self, other):
        """
        Two messages are equal if they have the same keys, and for each key the elements are equal.
        """
        tensors_equal = (self.tensor_message == other.tensor_message)
        df_equal = (self.df.equals(other.df))

        return tensors_equal and df_equal

    def __getitem__(self, index):
        """

        Args:
            index: An integer, slice, or list of integer indices.

        Returns:
            item: The element of the Message referenced by the index.
        """

        t = type(index)
        if t is int:
            if index >= self.length:
                raise IndexError
            index = slice(index,index+1)
            return self._getindex(index)
        elif t is slice:
            start = index.start or 0
            stop = index.stop or self.length
            if max(start, stop) > self.length:
                raise IndexError
            return self._getindex(index)
        elif t is list: # Must account for list of indices and list of strings
            if len(index) == 0:
                return Message(length=0)
            tt = type(index[0])
            if tt is int: # Is index
                return self._getindex(index)
            if tt is str: # Is list of column names
                return Message({s:self.tensor_message[s] if s in self.tensor_message.keys() else self.df[s] for s in index}, length = slice_length(index))
        elif t is str:
            if index in self.tensor_message.keys():
                return self.tensor_message[index]
            elif index in self.df.keys():
                return self.df[index]
        else:
            raise KeyError("{0} is not a column name or a valid index for this message.".format(str(index)))

    def _getindex(self, index):
        """
        Returns a submessage based on requested index.

        Args:
            index: An integer, slice, or list of integer indices.

        """
        if len(self.tensor_message) == 0:
            return Message(self.df.iloc[index].reset_index(drop=True), length=slice_length(index))
        if len(self.df) == 0:
            return Message(self.tensor_message[index], length=slice_length(index))
        else:
            return Message(self.tensor_message[index], self.df.iloc[index].reset_index(drop=True), length=slice_length(index))

    def __setitem__(self, index, value):
        """

        Args:
            index: An integer, slice, or list of integer indices.
            value: The value to set the index to. Must be an object that can be fed to the Message initializer and produce a message
                with the same length as the index.
        """
        t = type(index)

        if t is str:
            # Check if the update would require moving a column from df to tensor_message or vice-versa, in which case, delete the original
            # and then insert the new value into the correct location.
            if (type(value) is torch.Tensor and index in self.df.columns) or (not type(value) is torch.Tensor and index in self.tensor_message.columns):
                del self[index]

            # Update or insert
            if type(value) is torch.Tensor:
                self.tensor_message[index] = value # This will trigger a length check automatically
            else:
                if len(value) == self.length: # Check value before updating df
                    self.df[index] = value
                elif self.length == 0: # Inserting into empty Message is allowed
                    self.df[index] = value
                    self.check_length()
                else:
                    raise ValueError("Cannot set value {0} of length {1} to column name {2} for message with length {3}. Lengths of all columns \
                must be the same.".format(value, len(value), index, self.length))

        if t is int:
            index = slice(index, index+1)
            value = Message(value)
            self.tensor_message[index] = value.tensor_message
            value.df.index = self.df.iloc[index].index
            self.df.iloc[index] = value.df
        if t is slice:
            value = Message(value)
            self.tensor_message[index] = value.tensor_message
            value.df.index = self.df.iloc[index].index
            self.df.iloc[index] = value.df
        elif t is list:
            if len(index):
                tt = type(index[0])
                if tt is str: # Update multiple columns at once with a dict
                    for column in index:
                        self[column] = value[column]
                elif tt is int:
                    value = Message(value)
                    self.tensor_message[index] = value.tensor_message
                    value.df.index = self.df.iloc[index].index
                    self.df.iloc[index] = value.df
                else:
                    raise IndexError("{0} is not a valid index.".format(index))
            else:
                raise ValueError("No index specified.")

    def __delitem__(self, index):
        """
        Checks if index is a string (column name) or index (row, range of rows) and deletes the appropriate part of the message.
        If index is a list, it can be a list of either column names or indices.

        Args:
            index: An integer, slice, or list of integer indices.

        """
        t = type(index)
        if t is str: # Delete column
            if index in self.tensor_message.tensor_dict.keys():
                del self.tensor_message.tensor_dict[index]
            elif index in self.df.columns:
                del self.df[index]
            else:
                raise KeyError("Column {0} not found in message.")
        elif t is int: # Delete single row
            if index >= self.length:
                raise IndexError("Cannot delete element {0} from message with length {1}".format(index, self.length))
            self._delindex(index)
        elif t is slice:
            self._delindex(index)
        elif t is list:
            if len(index): # Otherwise list is empty and do nothing
                tt = type(index[0])
                if tt is str: # Delete columns
                    for column in index:
                        if column in self.tensor_message.tensor_dict.keys():
                            del self.tensor_message.tensor_dict[column]
                        elif column in self.df.columns:
                            del self.tensor_message[column]
                        else:
                            raise KeyError("Column {0} not found in message.")
                elif tt is int: # Delete multiple indices
                    m = max(index)
                    if m > self.length:
                        raise IndexError("Cannot delete element {0} from message with length {1}.".format(m, self.length))
                    else:
                        self._delindex(index)
        else:
            raise IndexError("{0} is not a column(s) or index in this message.".format(str(index)))
        if len(self.tensor_message.keys()) == 0:
            self.tensor_message.length = 0

    def _delindex(self, index):
        """
        Deletes the given index after it has been validated by __delitem__

        Args:
            index: An integer, slice, or list of integer indices.

        """
        if len(self.df):
            self.df.drop(self.df.index[index], inplace=True)
            self.length = len(self.df)
        if len(self.tensor_message):
            del self.tensor_message[index]
            self.length = self.tensor_message.length

    def __contains__(self, key):
        """
        This checks if the given key is present as a column in this Message.
        """
        if type(key) is str:
            return key in self.df or key in self.tensor_message
        else:
            return False

    def __getattr__(self, name):
        """
        If name is the name of a column in this Message, then this returns that column. Otherwise, this calls __getattr__ on the internal
        DataFrame or Tensor_Message.
        """
        if name in self.__getattribute__('df').keys():
            return self.df.__getattribute__(name)
        elif name in self.__getattribute__('tensor_message').keys():
            return self.tensor_dict.__getattr__(name)
        else:
            raise AttributeError("This message has no attribute or column named {0}.".format(name))

    def __repr__(self):
        return "Message with \n Tensors: \n {0} \n Metadata: \n {1}".format(self.tensor_message, self.df)

    @property
    def columns(self):
        """
        Returns the names of columns in this Message
        """
        return self.df.columns.append(pd.Index(self.tensor_message.keys()))

    def keys(self):
        """
        Returns names of keys in this Message.
        Note: This is the same as self.columns
        """

        return self.columns

    @property
    def index(self):
        """
        Returns index for internal tensors
        """
        # NOTE: Index currently has no meaning, because messages are currently required to be indexed from 0:length
        return self.df.index

    def append(self, other):
        """
        Compines messages together.
        Should initialize other if not a message already.

        Args:
            other: The message to append to self. Must have the same keys as self so that in the resulting Message,
                every column continues to have the same length as needed.
        """

        other = Message(other)
        # Check if self is empty, and if so, replace it with other.
        if self == empty_message:
            return other

        appended_tensors = self.tensor_message.append(other.tensor_message)
        appended_df = self.df.append(other.df).reset_index(drop=True)
        return Message(appended_tensors, appended_df)

    def enable_gradients(self, columns=None):
        columns = columns or self.tensor_message.columns
        for c in columns:
            self[c].requires_grad=True

    def merge(self, other):
        """
        Combines messages horizontally by producing a message with the keys/values of both.

        Args:
            other: The message to merge with self. Must have different keys and the same length as self to ensure length consistencies.
                Alternatively, if either self or other have an empty TensorMessage or df, then they can be merged together safely as long
                as the resulting Message has a consistent length.
                For example:
                ::
                    message_a = Message({'a': [1,2,3]}) # This is essentially a DataFrame
                    message_b = Message({'b': torch.Tensor([1,2,3])}) # This is essentially a TensorMessage
                    message_c = Message_a.merge(message_b) # This works

        Returns:
            message: The concatenated Message containing columns from self and other.

        """
        other = Message(other)
        return Message({**self.tensor_message, **other.tensor_message}, {**self.df, **other.df})

    def map(self, mapping):
        """
        Applies function mapping to message. If mapping is a dict, then maps will be applied to the corresponding keys as columns, leaving
        columns not present in mapping untouched.
        In otherwords, mapping would be a dict of column_name:functions specifying the mappings.

        Args:

            mapping: Can either be a dict mapping column names to functions that should be applied to those columns, or a single function.
                In the latter case, the mapping function will be applied to every column.

        Returns:

            message: A Message with the column:value pairs produced by the mapping.

        """
        if type(mapping) is dict:
            return Message({key: mapping[key](self[key]) if key in mapping else self[key] for key in self.columns})
        else:
            return Message({key: mapping(self[key]) for key in self.columns})

    def tensors(self, keys = None):
        """
        Return tensors associated with message as a tensormessage.
        If keys are specified, returns tensors associated with those keys, performing conversions as needed.

        Args:
            keys: Keys to get. Default = None, in which case all tensors are returned as a TensorMessage.
                If columns corresponding to requested keys are not tensors, they will be converted.

        Returns:
            tensors (TensorMessage): A TensorMessage containing the tensors requested.

        """
        if keys is None:
            return self.tensor_message
        else:
            return Message(TensorMessage({key:self[key] for key in keys}))

    def to_tensors(self, keys=None): # TODO: Test
        """
        Returns message with columns indicated by keys converted to Tensors. If keys is None, all columns are converted.

        Args:
            keys: Keys to get. Default = None, in which case all columns are mapped to Tensor.

        Returns:
            message: A Message in which the desired columns are Tensors.
        """
        if keys is None:
            keys = list(self.df.keys())

        tensor_message = Message(TensorMessage({key:self[key] for key in keys}))
        self[keys] = tensor_message

        return self

    def dataframe(self, keys = None):
        """
        Returns message as a dataframe. If keys are specified, only returns those keys as a dataframe.

        Args:
            keys: Keys to get. Default = None, in which case all non-tensors are returned as a DataFrame.
                If columns corresponding to requested keys are tensors, they will be converted (to np.arrays).

        Returns:
            df (pd.DataFrame): A DataFrame containing the columns requested.

        """
        if keys is None:
            return self.df
        else:
            return pd.DataFrame({key: np.array(self[key]).tolist() for key in keys})

    def to_dataframe(self, keys = None):
        """
        Returns message with columns indicated by keys converted to DataFrame. If keys is None, all tensors are converted.

        Args:
            keys: Keys to get. Default = None, in which case all tensors are mapped to DataFrame.

        Returns:
            message: A Message in which the desired columns are DataFrames.
        """
        if keys is None:
            keys = list(self.tensor_message.keys())

        new = Message(self.df)

        for key in keys:
            tensor = self[key].cpu().detach().numpy()
            if len(tensor.shape) > 1:
                tensor = tensor.tolist()
            new[key] = tensor

        return new

    def permute(self, index):
        """
        Reorders elements of message based on index.

        Args:
            index: A valid index for the message.

        Returns:
            message: A new Message with the elements arranged according to the input index.
                For example,
                ::
                message_a = Message({'a':[1,2,3]})
                message_b = message_a.permute([2,1,0])
                message_c = Message({'a': [3,2,1]})
                message_b == message_c

                The last statement will evaluate to True
        """
        df = self.df
        tensor_message = self.tensor_message

        if len(df):
            df_index = self.df.index
            df = self.df.reindex(index)
            df.index = df_index
        if len(tensor_message):
            tensor_message = self.tensor_message.permute(index)

        return Message(tensor_message, df)

    def cpu(self, keys = None):
        """
        Moves tensors to system memory. Can specify which ones to move by specifying keys.

        Args:
            keys: Keys to move to system memory. Default = None, meaning all columns are moved.

        Returns:
            message (Message): Moved message
        """
        self.tensor_message = self.tensor_message.cpu(keys)

    def cuda(self, device = 0, keys = None):
        """
        Moves tensors to gpu with given device number. Can specify which ones to move by specifying keys.

        Args:
            device (int): CUDA device number to use. Default = 0.
            keys: Keys to move to GPU. Default = None, meaning all columns are moved.

        Returns:
            message (Message): Moved message

        """
        self.tensor_message = self.tensor_message.cuda(device, keys)

        return self

class TensorMessage:
    """
    A TensorMessage is a class for representing data meant for consumption by pytorch as a dictionary of tensors. It is analogous to a
    Pandas DataFrame, except all elements are PyTorch Tensors (or torch.nn.Parameter). This can be useful when working with PyTorch, as you
    can have your Models use column names to keep track of what each Tensor is for, and by parameterizing column names, you can make your
    Models somewhat generic (instead of taking x as an argument that must have a specific format, you take take a TensorMessage and act on
    columns that can be configured at runtime or in your script.)

    TensorMessages, along with DataFrames, constitute the components of a Message. We recommend using Messages rather than TensorMessages
    directly, as all of the functionality of TensorMessages is encapsulated by a Message, and this gives you the ability to seamlessly move
    data to and from Tensor format, so you can mix your deep learning tasks with other data science tasks that you would use Pandas for.
    """
    def __init__(self, message_dict = None, map_dict = None):
        """
        Initizes TensorMessage, performing type conversions as necessary.
        """
        self.tensor_dict = {}
        self.map_dict = map_dict

        if type(message_dict) is Message:
            self.tensor_dict = message_dict.tensor_message.tensor_dict
            self.length = message_dict.tensor_message.length
        elif type(message_dict) is TensorMessage:
            self.tensor_dict = message_dict.tensor_dict
            self.length = message_dict.length
        elif message_dict is None:
            self.tensor_dict = {}
            self.length = 0
        else:
            for key, value in message_dict.items():
                # Make value list like if not already. For example, for length-1 messages.
                if not hasattr(value, '__len__'):
                    value = [value]
                # Apply (optional) mappings. For example, to specify a torch.BytesTensor instead of a FloatTensor.
                if type(self.map_dict) is dict and key in map_dict:
                    value = self.map_dict[key](value)
                # Convert to tensor if not already
                if not isinstance(value, torch.Tensor):
                    try:
                        value = torch.Tensor(value)
                    except Exception as err:
                        raise TypeError(
                        ("Elements of a message must be torch tensors. Tried to convert value associated with key {0} " + \
                        "but got error: {1}").format(key, err))
                # Make sure is not a 0-dim tensor
                if value.shape == torch.Size([]):
                    value = torch.Tensor([value])

                self.tensor_dict[key] = value

            self.length = compute_length(self.tensor_dict)

    def __getitem__(self, index): # TODO: Add support for nonstandard slices (0:, :-k, etc.)

        t = type(index)
        if t is str:
            return self.tensor_dict[index] # This will raise a key error if column not present
        elif t is int:
            if index >= self.length:
                raise IndexError
            else:
                return TensorMessage({key: value[index] for key, value in self.tensor_dict.items()})
        elif t is slice:
            start = index.start or 0
            stop = index.stop or self.length
            if max(start, stop) > self.length:
                raise IndexError
            else:
                return TensorMessage({key: value[index] for key, value in self.tensor_dict.items()})

        elif t is list: # Check if list of column names or indices
            if len(index):
                tt = type(index[0])
                if tt is str:
                    return TensorMessage({key:self.tensor_dict[key] for key in index})
                elif tt is int:
                    return TensorMessage({key: value[index] for key, value in self.tensor_dict.items()})
            else: # Empty list
                return TensorMessage()
        else:
            raise KeyError("{0} is not a column name or a valid index for this message.".format(str(index)))

    def __setitem__(self, index, value):
        """
        Sets the rows specified by the index to the provided value, which can be a dict of Tensors/list-likes, or a Message/TensorMessage/DataFrame
        of the appropriate length. All non-tensor objects will be converted to Tensor if possible (and this will cause an error if it's not possible).
        """
        t = type(index)
        if t is str:
            # Check length
            if len(value) != self.length and self.length:
                raise ValueError("Cannot set column {0} to value {1} with length {2} for message with length {3}".format(index, value, len(value), self.length))
            if not type(value) is torch.Tensor:
                value = torch.Tensor(value)
            self.tensor_dict[index] = value
            if self.length == 0: # Update length if inserting into empty TensorMessage
                self.length = len(value)

        elif t is int or t is slice:
            value = TensorMessage(value)
            for key in self.tensor_dict:
                self.tensor_dict[key][index] = value[key] # BUG: Out of bound slice queries return an empty Message instead of raising an error.
        elif t is list:
            if len(index):
                tt = type(index[0])
                if tt is str: # List of column names
                    for column, tensor in value.items():
                        self[column] = value # This will trigger recursive call to __setitem__
                elif tt is int:
                    value = TensorMessage(value)
                    for key in self.tensor_dict:
                        self.tensor_dict[key][index] = value[key] # Tensors can be indexed by list
                else:
                    raise IndexError("{0} is not a valid index.".format(index))
            else: # Empty list
                raise ValueError("No index specified for setitem.")

    def __delitem__(self, index):
        """
        Deletes the rows corresponding to the provided index.
        """
        t = type(index)
        if t is str:
            del self.tensor_dict[index] # Will automatically throw a KeyError if not in the dict
        elif t is int:
            if index > self.length:
                raise IndexError
            self._delindex(index)
        elif t is slice:
            # if max(index.stop, index.start) > self.length:
            #     raise IndexError
            self._delslice(index)
        elif t is list:
            if len(index):
                tt = type(index[0])
                if tt is str:
                    for column in index:
                        del self.tensor_dict[column]
                elif tt is int:
                    self._dellist(index)
            else:
                raise KeyError

        if len(self.keys()) == 0: # If you've deleted the last column, the length is now 0 #TODO: Test this
            self.length = 0

    def _delindex(self, index):
        """
        Deletes single row corresponding to a single index. This is called by __delitem__.
        """
        new_length = self.length - 1
        if index == 0:
            if self.length == 1: # Deleting the only row of a 1-row Message
                self.tensor_dict = {}
                self.length = new_length
            else: # Deleting first row of an n-row message

                self.tensor_dict = self[1:self.length].tensor_dict
                self.length = new_length
        elif index == self.length-1: # Deleting the last row
            self.tensor_dict = self[:self.length-1].tensor_dict
            self.length = new_length
        else: # Delete an in-between row
            self.tensor_dict = TensorMessage(cat([self[0:index], self[index+1:self.length]])).tensor_dict

    def _delslice(self, orange):
        """
        Deletes rows corresponding to a slice object. This is called by __delitem__.
        This involves finding the complementary set of indices and returning those.
        """
        if orange.step is None or orange.step in [0, 1, -1]: # Unit step size
            # NOTE: We will treat -1 and 1 as the same for now
            if orange.step == -1: # Is backwards
                start = orange.stop
                stop = orange.start
            else:
                start = orange.start
                stop = orange.step

            if start is None:
                start = 0
            if stop is None:
                stop = self.length

            if start >= stop:
                raise IndexError("Invalid slice {0}.".format(str(orange)))
            if start > 0:
                bottom_message = self[0:start]
            else:
                bottom_message = TensorMessage()
            if stop < self.length and stop > start:
                top_message = self[stop:self.length]
            else:
                top_message = TensorMessage()

            # TODO: Have to flip response if a downward slice is requested
            new_message = TensorMessage(cat([bottom_message, top_message]))

            self.tensor_dict = new_message.tensor_dict
            self.length = new_message.length

        else: # Step count is some number that must be accounted for
            listy = slice_to_list(orange) # NOTE: This is low priority so for now we just convert to list and delete that way
            self._dellist(listy)

    def _dellist(self, listy):
        """
        Deletes rows corresponding to indices in a list. This is called by __delitem__.
        This involves finding the complementary set of indices and returning those.
        """
        complement = [i for i in range(self.length) if i not in listy]
        self.tensor_dict = self[complement]
        self.length = len(complement)

    def __copy__(self):

        return TensorMessage(self.tensor_dict, self.map_dict)

    def __deepcopy__(self, memo):

        return TensorMessage(deepcopy(self.tensor_dict), deepcopy(self.map_dict))

    def __contains__(self, key):
        """
        Returns true if the requested key is a column in this TensorMessage.
        """
        if isinstance(key, Hashable):
            return key in self.tensor_dict
        else:
            return False

    def __getattr__(self, name):
        return self.tensor_dict.__getattribute__(name)

    def keys(self,*args, **kwargs):
        """
        Returns the column names of this TensorMessage, which as the keys of the internal tensor_dict.
        """
        return self.tensor_dict.keys(*args, **kwargs)

    def __len__(self):
        """
        The length of all columns in the TensorMessage must be the same to ensure consistency. By length, we mean the first dimension of a
        Tensor's shape (so if a Tensor is 2x4x6, its length is 2, and each 'row' consists of a 4x6 matrix.)
        """
        return self.length

    def __eq__(self, other):
        """
        Two tensor messages are equivalent if they have the same keys and their elements are equal.
        """

        keys_equal = set(self.keys()) == set(other.keys())
        if not keys_equal:
            return False
        for key in self.keys():
            if not torch.equal(self[key], other[key]):
                return False
        return True

    def __repr__(self):
        return "TensorMessage: {0}".format(self.tensor_dict)

    @property
    def tensor_message(self):
        return self

    @property
    def df(self):
        return pd.DataFrame()

    @property
    def columns(self):
        """
        Returns names of tensors in TensorMessage
        """
        return self.tensor_dict.keys()

    @property
    def index(self):
        """
        Returns the index for internal tensors
        """
        # NOTE: Index currently has no meaning, because messages are currently required to be indexed from 0:length
        return pd.RangeIndex(0,self.length)

    def enable_gradients(self, columns=None):
        columns = columns or self.columns
        for c in columns:
            self[c].requires_grad=True

    def disable_gradients(self, columns=None):
        pass

    def append(self, other):
        """
        Note if the target message has additional keys, those will be dropped.
        The target message must also have every key present in this message in
        order to avoid an value error due to length differences.
        """
        other = TensorMessage(other)

        # If self is empty, just replace it with other.
        if self == empty_tensor_message:
            return other
        return TensorMessage({key: torch.cat([self[key], other[key]]) if key in other.keys() else self[key] for key in self.tensor_dict.keys()})

    def merge(self, other):
        """
        Combines self with other into a message with the keys of both.
        self and other must have distinct keys.
        """
        other = TensorMessage(other)
        return TensorMessage({**self, **other})

    def permute(self, index):
        """
        Rearranges elements of TensorMessage to align with new index.
        """
        permutation = self[index]
        return permutation

    def cuda(self, device = 0, keys=None):
        """
        Moves all tensors in TensorMessage to cuda device specified by device number. Specify keys to limit the transformation
        to specific keys only.
        """
        if keys is None:
            keys = self.tensor_dict.keys()
        for key in keys:
            self.tensor_dict[key] = self.tensor_dict[key].cuda(device)

        return self

    def cpu(self, keys=None):
        """
        Moves all tensors in TensorMessage to cpu. Specify keys to limit the transformation
        to specific keys only.
        """
        if keys is None:
            keys = self.tensor_dict.keys()
        for key in keys:
            self.tensor_dict[key] = self.tensor_dict[key].cpu()

        return self

def compute_length(of_this):
    """
    Of_this is a dict of listlikes. This function computes the length of that object, which is the length of all of the listlikes, which
    are assumed to be equal. This also implicitly checks for the lengths to be equal, which is necessary for Message/TensorMessage.
    """
    lengths = [len(value) for value in of_this.values()]
    if lengths == []:
        # An empty dict has 0 length.
        return 0
    if len(set(lengths)) != 1: # Lengths are not all the same
        raise ValueError("Every element of dict must have the same length.")

    return lengths[0]

def extract_tensors(from_this):
    """
    Given a dict from_this, returns two dicts, one containing all of the key/value pairs corresponding to tensors in from_this, and the other
    containing the remaining pairs.
    """
    tensor_dict = {k:v for k, v in from_this.items() if type(v) is torch.Tensor}
    other_keys = from_this.keys() - tensor_dict.keys() if tensor_dict else from_this.keys()
    other_dict = {k: from_this[k] for k in other_keys}
    return tensor_dict, other_dict

def slice_to_list(s):
    """
    Converts a slice object to a list of indices
    """
    step = s.step or 1
    start = s.start
    stop = s.stop
    return [x for x in range(start,stop,step)]

def complement(indices, n):
    """ Given an index, returns all indices between 0 and n that are not in the index. """

    if type(indices) is slice:
        indices = slice_to_list(indices)
    if type(indices) is int:
        indices = [indices]

    complement = [i for i in range(n) if i not in indices]

    return complement

def cat(list_of_args):
    """
    Concatenates messages in list_of_args into one message.
    """

    if list_of_args == []:
        return Message()
    m = Message(list_of_args[0])
    if len(list_of_args) > 1:
        for arg in list_of_args[1:]:
            m = m.append(Message(arg))

    return m

def merge(list_of_args):
    """
    Merges messages in list_of_args into one message with all the keys combined.
    """

    m = Message(list_of_args[0])
    if len(list_of_args) > 1:
        for arg in list_of_args[1:]:
            m = m.merge(Message(arg))

    return m

# These are here for comparison purposes when testing if something is empty and is used by some of the methods in this file.
empty_tensor_message = TensorMessage()
empty_message = Message()



from abc import ABC, abstractmethod
from Bio import SeqIO
import pandas as pd
import fireworks
from fireworks.core import Pipe, HookedPassThroughPipe, recursive
from fireworks.core import Message
from fireworks.utils import index_to_list
from fireworks.core.cache import LRUCache, LFUCache, UnlimitedCache
from fireworks.utils import one_hot
from abc import ABC, abstractmethod
from itertools import count
import types
import random
from bidict import bidict
import torch
import numpy as np

class BioSeqPipe(Pipe):
    """
    Class for representing biosequence data.
    Specifically, this class can read biological data files (such as fasta) and iterate through them as a Pipe.
    This can serve as the first Pipe in a pipeline for analyzing genomic data.
    """

    name = 'BioSeqPipe'

    def __init__(self, path, input = None, filetype = 'fasta', **kwargs):
        """
        Args:
            path: Path on disk where file is located; will be supplied to the SeqIO.parse function.
            inputs: Input Pipes. Default is None.
            filetype: Type of file that will be supplied to the SeqIO.parse function. Default is 'fasta'
            kwargs: Optional key word arguments that will be supplied to the SeqIO.parse function.
        """
        Pipe.__init__(self, input=input)
        self.path = path
        self.filetype = filetype
        self.kwargs = kwargs
        self.seq = SeqIO.parse(self.path, self.filetype, **self.kwargs)

    @recursive
    def reset(self):
        """
        Resets the iterator to the beginning of the file. Specifically, it calls SeqIO.parse again as if reinstantiating the object.
        """
        self.seq = SeqIO.parse(self.path, self.filetype, **self.kwargs)

    def __next__(self):
        """
        Each iteration through a SeqIO object will produce a named tuple. This Pipe converts that tuple to a Message and produces the result.
        """
        gene = self.seq.__next__()

        try:
            return Message({
                'sequences': [str(gene.seq)],
                'ids': [gene.id],
                'names': [gene.name],
                'descriptions': [gene.description],
                'dbxrefs': [gene.dbxrefs],
            })
        except StopIteration:
            raise StopIteration

    def __iter__(self):

        self.reset()
        return self

class LoopingPipe(Pipe):
    """
    This Pipe can take any iterator and make it appear to be indexable by iterating through the input
    as needed to reach any given index.

    The input Pipe must implement __next__ and reset (to be repeatable),
    and this will simulate __getitem__ by repeatedly looping through the iterator as needed.

    For example, say we have a Pipe that iterates through the lines of a FASTA file:
    ::

        fasta = BioSeqPipe('genes.fasta')

    This Pipe can only iterate through the file in one direciton. If we want to access arbitrary elements,
    we can do this:
    ::

        clock = LoopingPipe(inputs=fasta)
        clock[10]
        clock[2:6]
        len(clock)

    All of these actions are now possible. Note that this is in general an expensive process, because the Pipe
    has to iterate one at a time to get to the index it needs. In practice, this Pipe should pipe its output
    to a CachingPipe that can store values in memory. This approach enables you to process datasets that don't
    entirely fit in memory; you can stream in what you need and cache portions. From the perspective of the downstream
    Pipes, every element of the dataset is accessible as if it were in memory.

    """

    name = 'LoopingPipe'

    def __init__(self, input, *args, **kwargs):
        super().__init__(input, *args, **kwargs)
        self.check_input()
        self.reset()
        self.length = None

    def __getitem__(self, index):
        """
        Retreives items in index by looping through inputs as many times as needed.
        """

        # TODO: Check if index exceeds length, either explicitly or implicitly.

        # Sort index
        index = sorted(index_to_list(index))
        above = [i for i in index if i >= self.position] # Go forward to reach these
        below = [i for i in index if i < self.position] # Will have to reset the loop to reach these
        if len(above) > 0:
            above_values = fireworks.cat([self.step_forward(i+1) for i in above])
        else:
            above_values = Message()
        if len(below) > 0:
            self.reset() # Position will now be reset to 0
            below_values = fireworks.cat([self.step_forward(i+1) for i in below])
        else:
            below_values = Message()
        return below_values.append(above_values) # TODO: Resort this message so values are in the order requested by index

    def __len__(self):
        """
        Length is also calculated implicitly as items are accessed.
        """
        if self.length is not None:
            return self.length
        else:
            self.compute_length()
            return self.length

    def __next__(self): # QUESTION: Should __next__ calls be supported for this Pipe?
        """
        Go forward one step. This can be used to loop over the inputs while automatically recording
        length once one cycle has been performed.
        """
        if self.length is None:
            return self.step_forward(1)
        else:
            p = (self.position + 1) % self.length
            self.position = p
            return self[p]

    def check_input(self):
        """
        Checks input to determine if it implements __next__ and reset methods.
        """
        # for name, pipe in self.inputs.items():
        pipe = self.input

        if pipe is None:
            raise TypeError("This pipe has no inputs.")
        if not (hasattr(pipe, '__next__') and hasattr(pipe, 'reset')):
            raise TypeError('Pipe {0} does not have __next__ and reset methods.'.format(name))

    @recursive()
    def reset(self):

        self.position = 0

    def compute_length(self):
        """
        Step forward as far as the inputs will allow and compute length.

        Length is also calculated implicitly as items are accessed.

        Note: If the inputs are infinite, then this will go on forever.
        """
        while True:
            try:
                self.step_forward(self.position+1)
            except (StopIteration, IndexError):
                self.reset()
                break

    def step_forward(self, n):
        """
        Steps forward through inputs until position = n and then returns that value.

        This also updates the internal length variable if the iterator ends due to this method call.
        """
        if self.length is not None and n > self.length:
            raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))
        if n < self.position:
            raise IndexError("Can only step forward to a value higher than current position.")
        x = Message()
        for _ in range(n - self.position):
            try:
                # x = x.append(fireworks.merge([Pipe.__next__() for Pipe in self.inputs.values()]))
                # x = fireworks.merge([pipe.__next__() for pipe in self.inputs.values()])
                x = self.input.__next__()
                self.position += 1
            except StopIteration:
                self.length = self.position
                raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))
        return x

class CachingPipe(Pipe):
    """
    This Pipe can be used to dynamically cache elements from upstream Pipes.
    Whenever data is requested by index, this Pipe will intercept the request and add that message alongside
    the index to its internal cache.
    This can be useful for dealing with datasets that don't fit in memory or are streamed in. You can cache portions
    of the dataset as you use them. By combining this with a LoopingPipe, you can create the illusion of making
    the entire dataset available to downstream Pipes regardless of the type and size of the original data.

    More specifically, fiven input Pipes that implement __getitem__, will store all calls to __getitem__ into
    an internal cache and therafter __getitem__ calls will either access from the cache or trigger __getitem__ calls
    on the input and an update to the cache.

    For example,

    ::

        fasta = BioSeqPipe(path='genes.fasta')
        clock = LoopingPipe(inputs=fasta)
        cache = CachingPipe(inputs=clock, cache_size=100) # cache_size is optional; default=100

    Will set up a pipeline that reads lines from a FASTA file and acessess and caches elements as requests are made

    ::

        cache[20:40] # This will be cached
        cache[25:30] # This will read from the cache
        cache[44] # This will update the cache
        cache[40:140] # This will fill the cache, flushing out old elements
        cache[25:30] # This will read from the dataset and update the cache again

    """

    def __init__(self, input, *args, cache_size = 100, buffer_size = 0, cache_type = 'LRU', infinite = False, **kwargs):
        """
        Args:
            args: Positional arguments for superclass initialization.
            cache_size (int): Number of elements to hold in cache. Default = 100
            buffer_size (int): Number of elements to clear once cache is full. This is useful if you expect frequent cache misses and
                don't expect to reuse previous cached elements. Default = 0
            cache_type (str): Type of cache. Choices are 'LRU' (least-recently-used) and 'LFU' (least-frequently-used). Default = 'LRU'
            infinite (bool): If set to true, the cache will have no upper limit on size and will never clear itself.
        """
        super().__init__(input, *args, **kwargs)
        self.check_input()
        self.length = None
        self.lower_bound = 0
        self.upper_bound = None
        self.infinite = infinite
        self.cache_size = cache_size
        self.cache_type = cache_type
        self.buffer_size = buffer_size
        self.init_cache(*args, **kwargs)


    def init_cache(self, *args, **kwargs):
        """
        This initializes a cache object at self.cache. There are currently two types of Cache available; LRUCache and LFUCache, and you can
        choose which one by specifying the cache_type argument in the initializer. See fireworks/core/cache.py for more information on
        Message caches.
        """
        choices = {'LRU': LRUCache, 'LFU': LFUCache}
        self.cache = choices[self.cache_type](max_size = self.cache_size, buffer_size=self.buffer_size)

    def check_input(self):
        """
        Checks inputs to determine if they implement __getitem__.
        """
        pipe = self.input
        if pipe is None:
            raise AttributeError("This pipe has no inputs.")

        if not (hasattr(pipe, '__getitem__')):
            raise TypeError('Pipe {0} does not have __getitem__ method.'.format(name))

    def __getitem__(self, index):

        index = index_to_list(index)
        if self.length and len(index): # Implicit length check if length is known
            if max(index) >= self.length:
                raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))

        # Identify what is in the cache and what isn't.
        in_cache = [i for i in index if i in self.cache.pointers] # Elements of index corresponding to in_cache elements
        in_cache_indices = [j for i,j in zip(index, count()) if i in self.cache.pointers] # Indices in index corresponding to in_cache elements
        not_in_cache = [i for i in index if i not in self.cache.pointers] # Elements of index corresponding to not_in_cache elements
        not_in_cache_indices = [j for i,j in zip(index, count()) if i not in self.cache.pointers] # Indices in index corresponding to not_in_cache elements
        # Retrieve from cache existing elements
        in_cache_elements = self.cache[in_cache] # elements in cache corresponding to indices in cache
        # Update cache to have other elements
        not_in_cache_elements = self.input[not_in_cache]
        self.cache[not_in_cache] = not_in_cache_elements
        # Reorder and merge requested elements
        message = in_cache_elements.append(not_in_cache_elements)
        indices = in_cache_indices
        indices.extend(not_in_cache_indices)
        # Elements must be reordered based on their order in index
        permutation = indices.copy()
        for i,j in zip(indices, count()):
            permutation[i] = j
        message = message.permute(permutation)
        # Implicit update of internal knowledge of length
        if len(index) and self.length is None and not self.infinite:
            l = max(index)
            if l > self.lower_bound:
                self.lower_bound = l

        return message

    def __len__(self):
        """
        Length is computed implicitly and lazily. If any operation causes the Pipe
        to reach the end of it's inputs, that position is stored as the length.
        Alternatively, if this method is called before that happens, the Pipe will attempt to
        loop to the end and calculate the length.
        """
        if self.infinite:
            raise ValueError("Pipe is labelled as having infinite length (ie. yields items indefinitely).")
        elif self.length is None:
            self.compute_length()
            return self.length
        else:
            return self.length

    def compute_length(self):
        """
        Step forward as far as the inputs will allow and compute length.
        Note: If the inputs are infinite, then this will go on forever.
        """
        if not self.infinite:
            while True:
                try:
                    self[1 + self.lower_bound] # Use knowledge of lower bound from previous getitem calls to accelerate this process
                except IndexError:
                    self.length = 1 + self.lower_bound
                    break
        else:
            raise ValueError("Pipe is labelled as having infinite length (ie. yields items indefinitely).")



class Title2LabelPipe(HookedPassThroughPipe):
    """
    This Pipe takes one Pipe as input and inserts a column called 'label'
    containing the provided title of the input Pipe to to all outputs.
    """

    def __init__(self, title, input, *args, labels_column = 'labels', **kwargs):

        super().__init__(input, *args, **kwargs)
        self.labels_column = labels_column
        self.label = title

    def _get_item_hook(self, message):

        return self.insert_labels(message)

    def _next_hook(self, message):

        return self.insert_labels(message)

    def insert_labels(self, message):

        l = len(message)
        message[self.labels_column] = [self.label for _ in range(l)]

        return message

class LabelerPipe(Pipe):
    """
    This Pipe implements a to_tensor function that converts labels contained in messages to tensors based on an internal labels dict.
    """

    def __init__(self, input, labels, *args, **kwargs):
        super().__init__(input, *args, **kwargs)
        self.labels = labels
        self.labels_dict = bidict({label:i for label,i in zip(labels,count())})

    def to_tensor(self, batch, labels_column = 'labels'):

        batch['labels_name'] = batch[labels_column]
        labels = batch[labels_column]
        labels = [self.labels_dict[label] for label in labels]
        labels = torch.Tensor([one_hot(label,len(self.labels_dict)) for label in labels])
        batch['labels'] = labels
        try: # Allow an upstream Pipe to attempt to perform to_tensor if it can
            batch = self.recursive_call('to_tensor', batch)
            return batch
        except AttributeError:
            return batch

class RepeaterPipe(Pipe):
    """
    Given an input Pipe that is iterable, enables repeat iteration. In other words, one loop through a RepeaterPipe is equivalent to n loops
    through the original dataset, where n is the number of repetitions that have been configured. Thsi can be useful for oversampling a data
    set without having to duplicate it.
    """

    def __init__(self, input, *args, repetitions=10,**kwargs):
        super().__init__(input, *args,**kwargs)
        if not type(repetitions) is int:
            raise ValueError("Number of repetitions must be provided as an integer.")

        self.repetitions = repetitions
        if not isinstance(self.input, Pipe): # TODO: Test this scenario
            self.iterator = iter(input)

    @recursive()
    def reset(self):
        self.iteration = 0

        if not isinstance(self.input, Pipe):
            self.iterator = iter(self.input)

    def __iter__(self):

        self.reset()
        return self

    def __next__(self):

        if not isinstance(self.input, Pipe):
            return self.iterator.__next__()
        else:
            try:
                return self.recursive_call('__next__')
            except StopIteration:
                self.iteration += 1
                if self.iteration >= self.repetitions:
                    raise StopIteration
                else:
                    self.recursive_call('reset')
                    return self.__next__()

class ShufflerPipe(Pipe):
    """
    Given input Pipes that implement __getitem__ and __len__, will shuffle the indices so that iterating through
    the Pipe or calling __getitem__ will return different values.
    """

    def __init__(self, input, *args, **kwargs):
        super().__init__(input, *args, **kwargs)
        self.check_input()
        self.current_index = 0
        self.shuffle_indices = np.array([i for i in range(len(self))])

    def check_input(self):
        """
        Check inputs to see if they implement __getitem__ and __len__
        """
        # for name, pipe in self.inputs.items():
        pipe = self.input
        if pipe is None:
            raise TypeError("This Pipe has no inputs.")
        if not (hasattr(pipe, '__getitem__')):
            raise TypeError('Pipe {0} does not have __getitem__ method.'.format(name))
        if not(hasattr(pipe, '__len__')):
            raise TypeError('Pipe {0} does not have __len__ method.'.format(name))

    def __getitem__(self, index):
        new_index = self.shuffle_indices[index]
        return self.recursive_call('__getitem__', new_index.tolist())

    def __next__(self):

        if self.current_index >= len(self):
            raise StopIteration

        message = self[self.current_index]
        self.current_index += 1
        return message

    def shuffle(self, order = None): # TODO: Add support for ordering

        np.random.shuffle(self.shuffle_indices)

    @recursive()
    def reset(self):

        self.shuffle()
        self.current_index = 0

    def __iter__(self):
        self.reset()
        return self

class IndexMapperPipe(Pipe):
    """
    Given input Pipes that implement __getitem__, returns a Pipe that maps indices in input_indices to output_indices via __getitem__
    """
    def __init__(self, input_indices, output_indices, *args,**kwargs):
        super().__init__(*args, **kwargs)
        self.check_input()
        if len(input_indices) != len(output_indices):
            raise ValueError("The number of input indices does not match the number of output indices.")
        self.pointers = UnlimitedCache()
        self.pointers[input_indices] = Message({'output':output_indices})
        self._current_index = 0

    def check_input(self):
        # for name, pipe in self.inputs.items():
        pipe = self.input
        if pipe is None:
            raise TypeError("This Pipe has no inputs.")
        if not hasattr(pipe, '__getitem__'):
            raise ValueError("Input Pipes must be indexable.")

    def __iter__(self):

        self._current_index = 0
        return self

    def __next__(self):

        try:
            iteration = self[self._current_index]
            self._current_index += 1
            return iteration
        except (KeyError, IndexError) as e:
            raise StopIteration

    def __getitem__(self, index):

        index = index_to_list(index)

        return self.input[self.pointers[index]['output'].values.tolist()]

    def __len__(self):
        return len(self.pointers)

class BatchingPipe(Pipe):
    """
    Generates minibatches.
    """

    def __init__(self, *args, batch_size=5, **kwargs):

        super().__init__(*args,**kwargs)
        self.current_index = 0
        self.batch_size = batch_size

    def __iter__(self):

        self.reset()
        return self

    @recursive()
    def reset(self):

        self.current_index = 0

    def __next__(self):

        if self.current_index + self.batch_size > len(self):
            raise StopIteration

        try:
            batch = self[self.current_index:self.current_index+self.batch_size]
            self.current_index += self.batch_size
            return batch
        except IndexError:
            raise StopIteration

class FunctionPipe(HookedPassThroughPipe):
    """
    This Pipe is initialized with a function that is applied to all instances of __getitem__, __next__, and __call__.
    This can be useful for quickly inserting a function into the middle of a pipeline.
    """
    def __init__(self, *args, function, **kwargs):
        super().__init__(*args, **kwargs)
        self._function = function

    def _call_hook(self, message): return self._function(message)

    def _next_hook(self, message): return self._function(message)

    def _getitem_hook(self, message): return self._function(message)

class TensorPipe(HookedPassThroughPipe):
    """
    This Pipe converts Messages to tensors. You can specify which columns should be converted.
    """
    def __init__(self, *args, columns=None, cuda=True, device=0, **kwargs):

        super().__init__(*args, **kwargs)
        self.columns = columns
        self.cuda = cuda and torch.cuda.is_available()
        self.device = device

    def _hook(self, message):

        if self.cuda:
            return message.to_tensors(self.columns).cuda(device=self.device, keys=self.columns)
        else:
           return message.to_tensors(self.columns)


    def _call_hook(self, message): return self._hook(message)

    def _next_hook(self, message): return self._hook(message)

    def _getitem_hook(self, message): return self._hook(message)

class GradientPipe(HookedPassThroughPipe):
    """
    This Pipe sets requires_grad=True for Tensor Columns. This is necessary for computing gradients when training a model.
    You can specify which columns should be converted.
    """
    def __init__(self, *args, columns=None, cuda=True, device=0, **kwargs):

        super().__init__(*args, **kwargs)
        self.columns = columns
        self.cuda = cuda and torch.cuda.is_available()
        self.device = device

    def _hook(self, message):

        message.enable_gradients(self.columns)
        return message

    def _call_hook(self, message): return self._hook(message)

    def _next_hook(self, message): return self._hook(message)

    def _getitem_hook(self, message): return self._hook(message)

"""
NOTE: Explore using the below code as an approach to make a general purpose wrapper Pipes,
ie. Pipes that modify all function calls from their input.
"""
# class GenericWrapperPipe(HookedPassThroughPipe):
#     """
#     Wraps all method and attribute calls with a hook.
#     """
#     def __getattr__(self, *args, **kwargs):
#         """
#         Pass through all methods of the input pipe while adding labels.
#         """
#         output = self.input.__getattribute__(*args, **kwargs)
#         if type(output) is types.MethodType: # Wrap the method in a converter
#             return self.method_wrapper(output)
#         else:
#             return self.attribute_wrapper(output)
#
#     def method_wrapper(self, function):
#         """
#         Wraps method with a label attacher such that whenever the method is called, the output is modified
#         by adding the label.
#         """
#
#         def new_function(*args, **kwargs):
#
#             output = function(*args, **kwargs)
#             try:
#                 output = Message(output)
#             except:
#                 return output
#
#             return self.insert_labels(output)
#
#         return new_function
#
#     def attribute_wrapper(self, attribute):
#         """
#         Wraps attribute with new label if attribute returns a message.
#         """
#         try:
#             output = Message(attribute)
#         except:
#             return attribute
#
#         return self.insert_labels(output)



from .message import Message
import os

class Scaffold:
    """
    A Scaffold can keep track of the internal state of objects in a pipeline. This can be used to save and load the entire state of a
    pipeline, allowing one to pause and resume a project, take snapshots, and log the internal state of components as an experiment proceeds.
    The current implementation of Scaffold is very simple; you attach objects to the Scaffold while providing a name that serves as an
    identifier for that object. You can then call the serialize method to get a dictionary of the current states of all attached objects,
    the save method to save those serialized states to a given folder, or load to update the states of attached objects using data in
    a provided directory.

    ::
    
        scaffold = Scaffold()
        # Attach components as desired
        scaffold.attach('a', A)
        scaffold.attach('b', B)
        .
        .
        .
        # This will save the current state of attached objects in folder 'save_directory'. The filenames will be based on the identifiers
        # eg. 'a.json', 'b.json', etc.
        scaffold.save(path='save_directory', method='json')
        # This will read files in folder 'save_directory2' and set the state of attached objects based on identifiers.
        # eg. the components attached with identifier 'a' will load from 'a.json' and so on. Note that the provided method must be
        # consistent with the filetypes in the save directory.
        scaffold.load(path='save_directory2', method='json')

        # This will produce a dictionary of identifiers to state object serializations (see Component_Map documentation for details).
        # You could use this dictionary to log state information however you want. For example, you could log the current weights of
        # neural network layers in your model for later plotting.

        state = scaffold.serialize()
    """

    def __init__(self, to_attach=None):
        """
        Args:
            to_attach: A dictionary of keys mapping to objects that you want the Scaffold to track. Note that each object must implement
                       a get_state method which returns a dictionary of the form {'external': {...}, 'internal': {...}}.
                       Pipes, Junctions, and Models satisfy this criteria.
        """
        self.attached = to_attach or {}

    def attach(self, name, obj):
        """
        Attaches an object to the Scaffold with a provided identifier. The Scaffold can then track the object's internal state, enabling one
        to access, save, and load the serialized states of all tracked objects at once.

        Args:
            name: The identifier for the object.
            obj: The object to attach. Note that each object must implement a get_state method which returns a dictionary of the
                 form {'external': {...}, 'internal': {...}}. Pipes, Junctions, and Models satisfy this criteria.
        """

        if not hasattr(obj, 'get_state'):
            raise AttributeError(
            "Objects attached to a Scaffold must have a 'get_state' method that returns a serialized representation of the object."
            )
        self.attached[name] = obj

    def serialize(self):
        """
        Returns a dictionary containing serialized representations of all objects tracked by the scaffold. See Component_Map documentation
        for more information on these serializations.

        Args:
            None

        Returns:
            state: A dict of the form {key: state}, where state is a dict of the form {'external': {...}, 'internal': {...}} corresponding to
            the internal and external state of objects tracked by the Scaffold. See Component_Map documentation for more information on state.
        """
        return {name: obj.get_state() for name, obj in self.attached.items()}

    def save(self, path, method='json', **kwargs):
        """
        Saves serialized representation of all objects linked to Scaffold using a desired method (json, csv, pickle, etc.)

        Args:
            path: The folder to save serializations to. This folder must exist and be writable by the program.
            method: The method for saving. This must be one of the methods support by the Message.to(...) method (see Message documentation)
                    , as state dicts are converted to Messages and saved using Message.to(...).
        """
        serialized = self.serialize()

        # Logic for saving
        for name, component_state in serialized.items():

            serialized_internal_state = Message.from_objects(component_state['internal'])
            component_path = os.path.join(path, '{0}.{1}'.format(name, method))
            kwargs['path'] = component_path
            saved_internal_state = serialized_internal_state.to(method=method, **kwargs)

    def load(self, path, method='json', reset=False):
        """
        Loads serialized representations of all objects linked to Scaffold using the given names in the given path.

        Args:
            path: The folder to load serializations from. This folder must exist and be readable by the program.
            method: The method for loading. This must be one of the methods support by the Message.load(...) method (see Message
                    documentation), as that method is used to read the files.
                    Note that load will only look for files corresponding to the provided method that also have the correspponding suffix
                    (eg. json filenames must end with '.json', pickles files with '.pickle', etc.). So if you have files in the foler that
                    were not saved as the given method, or have different filename suffixes, they will be ignored.
        """
        for filename in os.listdir(path):
            # state = Message.from_objects((os.path.join(path, filename))
            state = Message.read(method=method, path=os.path.join(path, filename)).to_dict()
            # NOTE: The saving process adds an extra dimension to the values, which we squash out here.
            state = {key: value[0] for key, value in state.items()}
            key = '.'.join(filename.split('.')[:-1]) # Drop the file suffix for mapping filenames to identifiers.

            if key in self.attached: # Files in the folder that do not correspond to attached objects will be ignored.

                self.attached[key].set_state({'external': {}, 'internal': state}, reset=reset)



from abc import ABC
from .message import Message
from fireworks.utils import index_to_list
from .cache import LRUCache, LFUCache, UnlimitedCache
from types import MethodType
from abc import ABC, abstractmethod

def recursive(accumulate=False):
    """
    Decorator that labels a Pipe method as recursive. This means, that method func will first be called on
    the Pipe's inputs and then on the Pipe itself.
    If accumulate is set to True, then the result from calling the method on a given Pipe will be
    used as input to the next one. If False, then the original arguments will be used when calling
    the method each time.
    """
    def outer(func):

        def wrapper(obj, *args, **kwargs):

            try:
                response = obj.input.__getattribute__(func.__name__)(*args, **kwargs)
            except AttributeError:
                response = None

            if accumulate:
                return func(obj, response)
            else:
                return func(obj, *args, **kwargs)

        return wrapper

    return outer

class Pipe(ABC):
    """
    The core object of computation in fireworks.
    A Pipe can take Pipes as inputs, and its outputs can be streamed to other Pipes.
    All communication is done via Message objects.
    Method calls are deferred to input Pipes recursively until a Pipe that implements the method is reached.

    This is made possible with a recursive function call method. Any Pipe can use this method to call a method on its inputs; this will recursively loop until reaching a Pipe that implements the method and return those outputs (as a Message) or raise an error if there are none. For example, we can do something like this:

    ::

        reader = pipe_for_reading_from_some_dataset(...)
        cache = CachingPipe(reader, type='LRU')
        embedder = CreateEmbeddingsPipe(cache})
        loader = CreateMinibatchesPipe(embedder})

        loader.reset()
        for batch in loader:
        	# Code for training

    Under the hood, the code for loader.__next__() can choose to recursively call a to_tensor() method which is implemented by embedder. Index queries and other magic methods can also be implemented recursively, and this enables a degree of commutativity when stacking Pipes together (changing the order of Pipes is often allowed because of the pass-through nature of recursive calls).

    Note that in order for this to work well, there must be some consistency among method names. If a Pipe expects ‘to_tensor’ to convert batches to tensor format, then an upstream Pipe must have a method with that name, and this should remain consistent across projects to maintain reusability. Lastly, the format for specifying inputs to a Pipe is a dictionary of Pipes. The keys in this dictionary can provide information for the Pipe to use or be ignored completely.

    """

    name = 'base_pipe'
    stateful_attributes = []

    def __init__(self, input = None, *args, **kwargs):

        self.input = input

    def __getitem__(self, *args, **kwargs):

        return self.input.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.input.__setitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        return self.input.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.input.__len__(*args, **kwargs)

    def __next__(self, *args, **kwargs):
        return self.input.__next__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.recursive_call('__call__', *args, **kwargs)

    def __iter__(self, *args, **kwargs):

        if hasattr(self.input, '__iter__'):
            self.input = self.input.__iter__(*args, **kwargs)
        elif hasattr(self.input, '__getitem__'):
            pass
        else:
            raise AttributeError("Input {0} to {1} is not iterable.".format(self.input, self))

        return self

    def __getattr__(self, name):
        """
        If the current pipe does not have the given attribute, this will recursively
        attempt to get the attribute from the input pipe.
        This will also not apply to attributes beginning with underscore. This way, Pipes can have local attributes and methods that do
        not clog up the pipeline namespace.
        This does not intercept special methods (__x__ methods)
        """

        if not name.startswith('_'):
            return self.recursive_call(name, call=False) #self.input.__getattribute__(*args, **kwargs)
        else:
            raise AttributeError("Pipe {0} has no attribute called {1}".format(self, name))

    def get_state(self):
        """
        This returns the current state of the Pipe, which consists of the values of all attributes designated in the list 'stateful_attributes'.
        This can be used to save and load a Pipe's state.

        Args:
            - None

        Returns:
            - A dict of the form {'internal': {...}, 'external': {...}}, where the 'external' subdict is empty. This is so that the representation
            is consistent with the get_state methods of Junctions and Models. We consider all attributes of a Pipe to be internal, and that is
            why the 'external' subdict is empty. See documentation on Component Map for more details on what we mean by that (note that Pipes
            don't use Component_Maps to store state, but simply expose similar methods for compatilibity.)
        """

        return {'internal': {key: getattr(self, key) for key in self.stateful_attributes}, 'external': {}}

    def set_state(self, state, *args, **kwargs):
        """
        Sets the state of the pipe based on the provided state argument.

        Args:
            - state: A dict of the form {'internal': {...}, 'external': {...}}. The 'external' dict will be ignored, because consider all
                     attributes of a Pipe to be in internal (for simplicity). See Component_Map documentation for details.
        """
        for key, value in {**state['internal']}.items():
            setattr(self, key, value)

    def recursive_call(self, attribute, *args, ignore_first = True, call=True,  **kwargs):
        """
        Recursively calls method/attribute on input until reaching an upstream Pipe that implements the method and
        returns the response as a message (empty if response is None).
        Recursive calls enable a stack of Pipes to behave as one entity; any method implemented by any component can be accessed
        recursively.

        Args:
            attribute: The name of the attribute/method to call.
            args: The arguments if this is a recursive method call.
            ignore_first: If True, then ignore whether or not the target attribute is implemented by self. This can be useful if a Pipe
                implements a method and wants to use an upstream call of the same method as well.
            call: If True, and the attribute is a method, the method will be called. Otherwise, it will be returned as is.
            kwargs: The kwargs is this is a recursive method call.

        Returns:
            Responses (dict): A dictionary mapping the name of each input Pipe to the response that was returned.
        """

        if not ignore_first:
            if hasattr(self, attribute):
                try:
                    attr = self.__getattribute__(attribute)
                    if type(attr) is MethodType and call:
                        return attr(*args, **kwargs)
                    else:
                        return attr

                except AttributeError:
                    return self.__getattr__(attribute)

        if not hasattr(self, 'input') or self.input is None:
            raise AttributeError("Pipe {0} does not have method/attribute {1}.".format(self.name, str(attribute)))

        if not isinstance(self.input, Pipe): # If input is not a pipe, just attempt a non-recursive method/attribute call on input.
            if args or kwargs and call: # Is a method call
                try:
                    return self.input.__getattribute__(attribute)(*args, **kwargs)
                except AttributeError:
                    raise AttributeError("Pipe {0} does not have method {1}.".format(self.name, str(attribute)))
            else: # Is an attribute
                try:
                    return self.input.__getattribute__(attribute)
                except AttributeError:
                    try:
                        return self.input.__getattr__(attribute)
                    except AttributeError:
                        raise AttributeError("Pipe {0} does not have attribute {1}".format(self.name, str(attribute)))

        response = self.input.recursive_call(attribute, *args, ignore_first=False, call=call, **kwargs)

        return response

    def _save_hook(self):
        """
        You can override this method to define custom behavior that taks place when a call to save() is made.
        """
        return {}

    def save(self, *args, **kwargs): # TODO: Implement save logic
        """
        This method currently does nothing but pass on a 'save' method call to the input.
        """
        save_dict = self._save_hook(*args, **kwargs)
        if save_dict == {}:
            pass
        else:
            save_df = Message.from_objects(save_dict).to_dataframe().df
            # Save the df using the given method and arguments
            # TODO: Implement

            # Save input

        self.input.save(*args, **kwargs)

class HookedPassThroughPipe(Pipe):
    """
    This Pipe has hooks which can be implemented by subclasses to modify the behavior of
    passed through calls.
    You can define hooks for the following (magic) methods: __getitem__, __call__, and __next__.
    Whenever you call one of these method this will happen:
        1. The method will be recursively called on this Pipes input (if it exists)
        2. The appropriate hook function will be called on the result of that recursive call
        3. This will be the returned value.

    These hooks can make it easy to create pipes that 'do something' every time data is accessed in a certain way. For example, you could have
    the pipe apply some transform to the data.
    """

    name = 'Hooked-passthrough Pipe'

    def _getitem_hook(self, message): return message

    def _call_hook(self, message): return message

    def _next_hook(self, message): return message

    def __getitem__(self, item): # TODO: wrap access methods in try/catch statements

        return self._getitem_hook(Message(self.input.__getitem__(item))) #self.input.__getitem__(*args, **kwargs))

    def __next__(self):
        return self._next_hook(Message(self.input.__next__()))

    def __call__(self, *args, **kwargs): # TODO: Test these hooks more thoroughly.

        try:
            return self._call_hook(self.recursive_call('__call__', *args, **kwargs))
        except AttributeError:
            return self._call_hook(*args, **kwargs)



import torch
from copy import deepcopy
from fireworks import Message
from fireworks.extensions import IgniteJunction
from fireworks.utils.exceptions import EndHyperparameterOptimization
from ignite.metrics import Metric
from ignite.exceptions import NotComputableError
from fireworks.extensions import LocalMemoryFactory, Experiment
from fireworks.extensions.training import default_training_closure, default_evaluation_closure
from itertools import combinations, count

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from nonlinear_regression_utils import NonlinearModel, get_data
from nonlinear_regression import base_loss, ModelSaverMetric

train_set, test_set, params = get_data(n=1000)
loss = lambda batch: base_loss(batch['y_pred'], batch['y'])

# Specify hyperparameter optimization scheme using Factory
def make_model(parameters):
    temp_parameters = deepcopy(parameters)
    include = [letter for letter in ['a','b','c','d','e'] if letter in parameters]
    exclude = [letter for letter in ['a','b','c','d','e'] if letter not in parameters]
    for letter in exclude:
        temp_parameters[letter] =  [0]
    model = NonlinearModel(temp_parameters)
    for letter in exclude: # Prevent training from taking place for these parameters
        model.freeze(letter)
    return model

def get_trainer(train_set, loss, optimizer, **kwargs):

    def train_from_params(parameters):

        model = make_model(parameters)
        trainer = IgniteJunction(components={'model': model, 'dataset': train_set}, loss=loss, optimizer=optimizer, visdom=False, **kwargs)
        print("Now training model for parameters {0}".format(parameters))
        trainer.train(max_epochs=10)
        evaluator = IgniteJunction(components={'model': model, 'dataset': train_set}, loss=loss, optimizer=optimizer, update_function=default_evaluation_closure, visdom=False, **kwargs)
        print("Now evaluating trained model.")
        return trainer

    return train_from_params

class Parameterizer:

    def __init__(self):
        possible_params = ['a','b','c','d','e']
        def generator():
            for i in reversed(range(5)):
                for combination in combinations(possible_params,i):
                    params = {param: [0] for param in combination}
                    yield params
        self.generator = generator()

    def __call__(self,past_params, metrics):
        try:
            params = self.generator.__next__()
            if params == {}:
                raise EndHyperparameterOptimization
            return params

        except StopIteration:
            raise EndHyperparameterOptimization

class AccuracyMetric(Metric):

    def __init__(self, output_transform = lambda x:x):
        Metric.__init__(self, output_transform=output_transform)
        self.reset()

    def reset(self):
        self.l2 = 0.
        self.num_examples = 0

    def update(self, output):
        self.l2 += output['loss']
        self.num_examples += len(output['output'])

    def compute(self):

        if self.num_examples == 0:
            raise NotComputableError(
                "Metric must have at least one example before it can be computed."
            )
        return Message({'average-loss': [self.l2 / self.num_examples]}).to_dataframe()

if __name__=="__main__":

    description = "In this experiment, we will compare the performance of different polynomial models when regressed against data generated from a random polynomial."
    experiment = Experiment("model_selection", description=description)

    factory = LocalMemoryFactory(components={
        'trainer': get_trainer(train_set, loss, optimizer='Adam', lr=.1),
        'eval_set': test_set,
        'parameterizer': Parameterizer(),
        'metrics': {'accuracy': AccuracyMetric(), 'model_state': ModelSaverMetric()}
        })

    factory.run()

    # Plot the different runs
    fig, ax = plt.subplots()
    model = NonlinearModel()
    true_model = NonlinearModel(components={'a':[params['a']], 'b': [params['b']], 'c': [params['c']], 'd': [0], 'e': [0]})
    x = Message({'x':np.arange(-10,10,.2)}).to_tensors()
    y_true = true_model(x)['y_pred'].detach().numpy()

    def animate(frame):

        current_state = {'internal': frame[1]['internal'][0], 'external': {}}
        model.set_state(current_state)

        y_predicted = model(x)['y_pred'].detach().numpy()
        xdata = list(x['x'].detach().numpy())
        ydata = list(y_predicted)
        ax.clear()
        ax.plot(xdata, list(y_true), 'r')
        ax.plot(xdata, ydata, 'g')

        # Set up titles
        substrings = []
        if (frame[0]['a'] == 0).all():
            substrings.append("a")
        if (frame[0]['b'] == 0).all():
            substrings.append("bx")
        for key, i in zip(['c','d','e'], count()):
            if (frame[0][key] == 0).all():
                substrings.append("{0}x^{1}".format(key,i+2))
        title = "Model: {0}".format(" + ".join(substrings))

        ax.set_title(title)

    params, metrics = factory.read()
    accuracy_file = experiment.open('accuracy.csv', string_only=True)
    metrics['accuracy'].to('csv', path=accuracy_file)
    model_state_file = experiment.open('model_states.csv', string_only=True)
    metrics['model_state'].to('csv', path=model_state_file)
    params_file = experiment.open('params.csv', string_only=True)
    params.to('csv', path=params_file)

    ani = FuncAnimation(fig, animate, zip(factory.params, factory.computed_metrics['model_state']), interval=3000)
    ani.save(experiment.open("models.mp4", string_only=True)) # This will only work if you have ffmpeg installed.
    plt.show()



from . import utils
import numpy as np

def test_one_hot():

    hot = utils.one_hot(2, 10)
    assert (hot == np.array([0,0,1,0,0,0,0,0,0,0])).all()

def test_subset_dict():

    d1 = {'a':1, 'b':2, 'c':3, 'd': 4}
    d2 = utils.subset_dict(d1, ['b', 'c', 'e'])
    assert d2['b'] == d1['b']
    assert d2['c'] == d2['c']
    assert 'e' not in d2

    d3 = utils.subset_dict(d1, ['f'])
    assert d3 == {}
